import pytest
import os
import sys
import copy
import flaky
from yggdrasil import tools, timing, platform
from tests import TestClassBase as base_class


_test_lang = 'c'
if 'c' not in timing.get_lang_list():  # pragma: windows
    _test_lang = 'python'  # pragma: testing
_this_platform = (platform._platform,
                  '%d.%d' % sys.version_info[:2],
                  tools.get_default_comm())
_base_environment = {'platform': 'Linux',
                     'python_ver': '2.7',
                     'comm_type': 'zmq'}
_valid_platforms = [('Linux', '2.7', 'zmq'),
                    ('Linux', '2.7', 'ipc'),
                    ('Linux', '3.5', 'zmq'),
                    ('MacOS', '2.7', 'zmq'),
                    ('Windows', '2.7', 'zmq')]
_testfile_json = 'test_run123.json'
_testfile_dat = 'test_run123.dat'


def test_get_source():
    r"""Test getting source file for test."""
    lang_list = timing.get_lang_list()
    dir_list = ['src', 'dst']
    for lang in lang_list:
        for d in dir_list:
            fname = timing.get_source(lang, d)
            assert(os.path.isfile(fname))


def test_platform_error():
    r"""Test error when test cannot be performed."""
    test_platform_map = {'MacOS': 'Linux',
                         'Linux': 'Windows',
                         'Windows': 'MacOS'}
    test_platform = test_platform_map[platform._platform]
    x = timing.TimedRun(_test_lang, _test_lang, platform=test_platform)
    with pytest.raises(RuntimeError):
        x.can_run(raise_error=True)


@pytest.mark.suite("timing", disabled=True)
class TimedRunTestBase(base_class):
    r"""Base test class for the TimedRun class."""

    _mod = 'yggdrasil.timing'
    _cls = 'TimedRun'

    @pytest.fixture
    def count(self):
        r"""int: Number of messages to use for tests."""
        return 1

    @pytest.fixture
    def size(self):
        r"""int: Size of messages to use for tests."""
        return 1

    @pytest.fixture
    def nrep(self):
        r"""int: Number of times to repeat."""
        return 1

    @pytest.fixture
    def language(self):
        r"""str: Language to test."""
        return _test_lang

    @pytest.fixture
    def instance_args(self, language):
        r"""Arguments for a new instance of the tested class."""
        return (language, language)

    @pytest.fixture
    def instance_kwargs(self):
        r"""Keyword arguments for a new instance of the tested class."""
        return {}

    @pytest.fixture
    def time_run_args(self, count, size):
        r"""tuple: Arguments for time_run."""
        return (count, size)

    @pytest.fixture
    def time_run_kwargs(self, nrep):
        r"""dict: Keyword arguments for time_run."""
        return {'nrep': nrep}

    @pytest.fixture
    def entry_name(self, instance, time_run_args):
        r"""str: Name of the entry for the provided time_run_args."""
        return instance.entry_name(*time_run_args)

    @pytest.fixture
    def filename(self, instance):
        r"""str: Name of the file where data is stored."""
        return instance.filename

    @pytest.fixture
    def check_filename(self, filename):
        r"""Raise a unittest.SkipTest error if the filename dosn't exist."""
        if not os.path.isfile(filename):  # pragma: debug
            pytest.skip(f"Performance stats file dosn't exist: {filename}")

    @pytest.fixture
    def get_raw_data(self, filename):
        r"""Get the raw contents of the data file."""
        def get_raw_data_w():
            out = ''
            if os.path.isfile(filename):
                with open(filename, 'r') as fd:
                    out = fd.read()
            return out
        return get_raw_data_w

    @pytest.fixture
    def time_run(self, instance, time_run_args, time_run_kwargs):
        r"""Perform a timed run."""
        def time_run_w(inst=instance):
            inst.time_run(*time_run_args, **time_run_kwargs)
        return time_run_w


@pytest.mark.production_run
class TestTimedRun(TimedRunTestBase):
    r"""Test class for the TimedRun class using existing data."""

    @pytest.fixture
    def instance_kwargs(self):
        r"""Keyword arguments for a new instance of the tested class."""
        return {'test_name': 'timed_pipe',
                'filename': None,
                'platform': _base_environment['platform'],
                'python_ver': _base_environment['python_ver'],
                'comm_type': _base_environment['comm_type'],
                'dont_use_pyperf': False,
                'max_errors': 5}

    @pytest.fixture
    def count(self, instance):
        r"""int: Number of messages to use for tests."""
        return instance.base_msg_count

    @pytest.fixture
    def size(self, instance):
        r"""int: Size of messages to use for tests."""
        return instance.base_msg_size

    @pytest.fixture
    def language(self):
        r"""str: Language to test."""
        return 'python'
    
    def test_json(self, check_filename, get_raw_data, instance):
        r"""Test loading/saving pyperf data as json."""
        old_text = get_raw_data()
        x = instance.load(as_json=True)
        instance.save(x, overwrite=True)
        new_text = get_raw_data()
        assert(new_text == old_text)

    def test_save(self, check_filename, get_raw_data, instance):
        r"""Test save with/without overwrite."""
        old_text = get_raw_data()
        with pytest.raises(RuntimeError):
            instance.save(instance.data)
        instance.save(instance.data, overwrite=True)
        new_text = get_raw_data()
        assert(new_text == old_text)

    def test_scaling_count(self, check_filename, instance, count, size,
                           nrep):
        r"""Test running scaling with number of messages."""
        kwargs = dict(min_count=count, max_count=count,
                      nsamples=1, nrep=nrep)
        instance.scaling_count(size, scaling='log', **kwargs)
        instance.scaling_count(size, scaling='linear',
                               per_message=True, **kwargs)
        with pytest.raises(ValueError):
            instance.scaling_count(size, scaling='invalid')

    def test_scaling_size(self, check_filename, instance, count, size,
                          nrep):
        r"""Test running scaling with size of messages."""
        kwargs = dict(min_size=size, max_size=size,
                      nsamples=1, nrep=nrep)
        instance.scaling_size(count, scaling='log', **kwargs)
        instance.scaling_size(count, scaling='linear',
                              per_message=True, **kwargs)
        with pytest.raises(ValueError):
            instance.scaling_size(count, scaling='invalid')

    @pytest.fixture
    def close_figures(self):
        r"""Close figures after the test."""
        yield
        import matplotlib.pyplot as plt  # noqa: E402
        plt.clf()
        plt.cla()
        plt.close('all')

    def test_plot_scaling_joint(self, check_filename, instance, count, size,
                                close_figures, disable_verify_count_fds):
        r"""Test plot_scaling_joint."""
        kwargs = dict(msg_size0=size, msg_count0=count,
                      msg_size=[size], msg_count=[count],
                      per_message=True, time_method='bestof')
        instance.plot_scaling_joint(**kwargs)

    def test_plot_scaling(self, check_filename, instance, count, size, nrep,
                          close_figures, disable_verify_count_fds):
        r"""Test plot_scaling corner cases not covered by test_plot_scaling_joint."""
        instance.plot_scaling(size, [count], per_message=True,
                              time_method='average', yscale='linear')
        instance.plot_scaling([size], count, per_message=False,
                              time_method='average', yscale='log')

        if False:
            # Test with msg_count on x linear/linear axs
            args = (size, [count])
            kwargs = {'axs': None, 'nrep': nrep,
                      'time_method': 'average', 'per_message': True}
            kwargs['axs'] = instance.plot_scaling(*args, **kwargs)
            kwargs['time_method'] = 'bestof'
            kwargs['axs'] = instance.plot_scaling(*args, **kwargs)
            # Test with msg_size on x log/log axes
            args = ([size], count)
            kwargs = {'axs': None, 'nrep': nrep,
                      'time_method': 'average',
                      'xscale': 'log', 'yscale': 'log'}
            kwargs['axs'] = instance.plot_scaling(*args, **kwargs)
            kwargs['time_method'] = 'bestof'
            kwargs['axs'] = instance.plot_scaling(*args, **kwargs)
        # Errors
        with pytest.raises(RuntimeError):
            instance.plot_scaling([size], [count])
        with pytest.raises(RuntimeError):
            instance.plot_scaling(size, count)
        with pytest.raises(ValueError):
            instance.plot_scaling([size], count, nrep=nrep,
                                  time_method='invalid')

    def test_pyperfjson_to_pandas(self, check_filename, filename):
        r"""Test pyperfjson_to_pandas."""
        timing.pyperfjson_to_pandas(filename)

    def test_fits(self, check_filename, instance):
        r"""Test fits to scaling on one platform."""
        instance.time_per_byte
        instance.time_per_message
        instance.startup_time

    def test_plot_scalings(self, check_filename, instance_kwargs,
                           count, size, language, close_figures,
                           disable_verify_count_fds):
        r"""Test plot_scalings corner cases on test platform."""
        kwargs = copy.deepcopy(instance_kwargs)
        kwargs.update(msg_size=[size], msg_size0=size,
                      msg_count=[count], msg_count0=count,
                      cleanup_plot=True)
        for c in ['comm_type', 'language', 'platform', 'python_ver']:
            ikws = copy.deepcopy(kwargs)
            ikws['compare'] = c
            if c in ikws:
                del ikws[c]
            if c == 'language':
                ikws['per_message'] = True
                ikws['compare_values'] = [language]
            timing.plot_scalings(**ikws)
        # Errors
        with pytest.raises(ValueError):
            timing.plot_scalings(compare='invalid')
        with pytest.raises(RuntimeError):
            timing.plot_scalings(compare='comm_type', comm_type='zmq')

    def test_production_runs(self, check_filename, instance_kwargs,
                             close_figures):
        r"""Test production tests (those used in paper)."""
        # Limit language list for tests
        for c in ['comm_type', 'language', 'platform', 'python_ver']:
            kwargs = copy.deepcopy(instance_kwargs)
            kwargs.update(compare=c, cleanup_plot=True, use_paper_values=True)
            if c in kwargs:
                del kwargs[c]
            timing.plot_scalings(**kwargs)
            # Also do MacOS plot w/ Matlab
            if c == 'language':
                kwargs['platform'] = 'MacOS'
                timing.plot_scalings(**kwargs)


class TestTimedRunTemp(TimedRunTestBase):
    r"""Test class for the TimedRun class using temporary data."""

    @pytest.fixture
    def instance_kwargs(self, dont_use_pyperf):
        r"""Keyword arguments for a new instance of the tested class."""
        if dont_use_pyperf:
            # This forces use of standard name with .dat extension
            filename = None
        else:
            filename = 'test_run123.json'
        return {'test_name': 'timed_pipe',
                'filename': filename,
                'platform': None,
                'python_ver': None,
                'comm_type': None,
                'dont_use_pyperf': dont_use_pyperf,
                'max_errors': 5}

    @pytest.fixture(scope="class", params=[False, True])
    def dont_use_pyperf(self, request):
        r"""Subtype of component being tested."""
        return request.param
    
    @pytest.fixture(autouse=True)
    def cleanup_files(self, instance):
        r"""Remove the temporary file if it exists."""
        yield
        if os.path.isfile(instance.filename):
            os.remove(instance.filename)
        if os.path.isfile(instance.pyperfscript):  # pragma: debug
            os.remove(instance.pyperfscript)

    @pytest.fixture
    def time_run_kwargs(self, nrep):
        r"""dict: Keyword arguments for time_run."""
        return {'nrep': nrep, 'overwrite': True}

    @flaky.flaky
    def test_pyperf_func(self, dont_use_pyperf, instance, count, size):
        r"""Test pyperf_func."""
        if dont_use_pyperf:
            pytest.skip("Don't use pyperf")
        timing.pyperf_func(1, instance, count, size, 0)

    @flaky.flaky
    def test_run_overwrite(self, time_run, python_class,
                           instance_args, instance_kwargs, entry_name):
        r"""Test performing a run twice, the second time with ovewrite."""
        time_run()
        # Reload instance to test load for existing file
        instance = python_class(*instance_args, **instance_kwargs)
        time_run(instance)
        instance.remove_entry(entry_name)
        assert(not instance.has_entry(entry_name))



import pytest
from tests.drivers.test_CompiledModelDriver import (
    TestCompiledModelDriver as base_class)
import os
import shutil


class TestBuildModelDriver(base_class):
    r"""Test parameters for BuildModelDriver."""

    parametrize_language = []
    
    test_build = None
    test_call_linker = None
    test_parse_arguments = None
    test_get_tool = None
    test_get_dependency_info = None
    test_get_dependency_source = None
    test_get_dependency_object = None
    test_get_dependency_library = None
    test_get_dependency_include_dirs = None
    test_get_dependency_order = None
    test_invalid_function_param = None
    test_compilers = None
    test_compile_model = None  # TODO: Verify
    test_get_linker_flags = None  # TODO: Verify
    
    @pytest.fixture
    def sourcedir(self, source):
        r"""Directory that source code is in."""
        return os.path.dirname(source[0])

    @pytest.fixture
    def target(self, source):
        r"""Make target that should be used."""
        return os.path.basename(source[0])
    
    @pytest.fixture
    def instance_args(self, name, target):
        r"""Arguments for a new instance of the tested class."""
        return (name, target)

    def test_get_language_for_source(self, python_class, source):
        r"""Test the get_language_for_source method."""
        buildfile = None
        if python_class.buildfile_base:
            buildfile = os.path.join(os.path.dirname(source[0]),
                                     python_class.buildfile_base)
            buildfile_cache = '_copy'.join(os.path.splitext(buildfile))
        python_class.get_language_for_source(source)
        python_class.get_language_for_source(os.path.dirname(source[0]))
        try:
            if buildfile:
                shutil.move(buildfile, buildfile_cache)
            python_class.get_language_for_source(source[0])
        finally:
            if buildfile:
                shutil.move(buildfile_cache, buildfile)



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
from yggdrasil.metaschema import MetaschemaTypeError
from yggdrasil.metaschema.datatypes.MultiMetaschemaType import (
    create_multitype_class)


class TestMultiMetaschemaType(base_class):
    r"""Test class for MultiMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.MultiMetaschemaType'
    _cls = 'MultiMetaschemaType'

    @pytest.fixture(scope="class")
    def types(self):
        r"""list: Test types."""
        return ['object', 'int']
    
    @pytest.fixture(scope="class", autouse=True)
    def python_class(self, types):
        r"""Python class that is being tested."""
        return create_multitype_class(types)

    @pytest.fixture(scope="class")
    def value(self):
        r"""dict: Test value."""
        return {'a': int(1), 'b': float(1)}
        
    @pytest.fixture(scope="class")
    def valid_encoded(self, typedef_base, types):
        r"""list: Encoded objects that are valid under this type."""
        return [dict(typedef_base, type=types)]

    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value, int(1)]

    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return ['hello']

    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]

    @pytest.fixture(scope="class")
    def typedef(self, typedef_base, types):
        r"""dict: Type definition"""
        return dict(typedef_base, type=types)

    def test_type_mismatch_error(self, python_class):
        r"""Test that error is raised when there is a type mismatch."""
        with pytest.raises(MetaschemaTypeError):
            python_class(type=['invalid'])



from yggdrasil import units, constants
from yggdrasil.metaschema import data2dtype, MetaschemaTypeError
from yggdrasil.metaschema.properties.MetaschemaProperty import (
    MetaschemaProperty)


class SubtypeMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'subtype' property."""

    name = 'subtype'
    schema = {'description': 'The base type for each item.',
              'type': 'string',
              'enum': [k for k in sorted(constants.VALID_TYPES.keys())]}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'subtype' scalar property."""
        dtype = data2dtype(instance)
        out = None
        for k, v in constants.VALID_TYPES.items():
            if dtype.name.startswith(v):
                out = k
                break
        if out is None:
            raise MetaschemaTypeError('Cannot find subtype string for dtype %s'
                                      % dtype)
        return out

    @classmethod
    def normalize_in_schema(cls, schema):
        r"""Normalization for the 'subtype' scalar property in a schema."""
        if cls.name in schema:
            return schema
        if not units.is_null_unit(schema.get('units', '')):
            schema.setdefault(cls.name, 'float')
        return schema
    

class PrecisionMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'precision' property."""
    
    name = 'precision'
    schema = {'description': 'The size (in bits) of each item.',
              'type': 'number',
              'minimum': 1}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'precision' scalar property."""
        dtype = data2dtype(instance)
        out = dtype.itemsize * 8  # in bits
        return out

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparison for the 'precision' scalar property."""
        if (prop1 > prop2):
            yield '%s is greater than %s' % (prop1, prop2)

    @classmethod
    def normalize_in_schema(cls, schema):
        r"""Normalization for the 'precision' scalar property in a schema."""
        if cls.name in schema:
            return schema
        subtype = schema.get('subtype', schema.get('type'))
        if subtype in ['float', 'int', 'uint']:
            schema.setdefault(cls.name, int(64))
        elif subtype in ['complex']:
            schema.setdefault(cls.name, int(128))
        return schema
            

class UnitsMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'units' property."""

    name = 'units'
    schema = {'description': 'Physical units.',
              'type': 'string'}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'units' scalar property."""
        out = units.get_units(instance)
        if (not out) and (typedef is not None):
            out = typedef
        return out

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparision for the 'units' scalar property."""
        if not units.are_compatible(prop1, prop2):
            yield "Unit '%s' is not compatible with unit '%s'" % (prop1, prop2)



import pytest
from yggdrasil.metaschema.datatypes.ContainerMetaschemaType import (
    ContainerMetaschemaType)


def test_container_errors():
    r"""Test implementation errors on bare container class."""
    with pytest.raises(NotImplementedError):
        ContainerMetaschemaType._iterate(None)
    with pytest.raises(NotImplementedError):
        ContainerMetaschemaType._assign(None, None, None)
    with pytest.raises(NotImplementedError):
        ContainerMetaschemaType._has_element(None, None)



import pytest
import os
from tests.examples import TestExample as base_class
from pandas.testing import assert_frame_equal
from yggdrasil.components import create_component


class TestExampleConditionalIO(base_class):
    r"""Test the conditional_io example."""

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "conditional_io"

    @pytest.fixture
    def expected_output_files(self, yamldir):
        r"""list: Examples of expected output for the run."""
        return [os.path.join(yamldir, 'Output', 'output.txt')]
        
    @pytest.fixture
    def output_files(self, yamldir):
        r"""list: Output files for the run."""
        return [os.path.join(yamldir, 'output.txt')]

    @pytest.fixture(scope="class")
    def read_file(self):
        r"""Read in contents from a file.

        Args:
            fname (str): Full path to the file that should be read.

        Returns:
            object: File contents.

        """
        def read_file_w(fname):
            x = create_component('file', 'table', name='test',
                                 address=fname, direction='recv',
                                 as_array=True, recv_converter='pandas')
            msg = x.recv_array()[1]
            if msg is not None:
                msg = msg.sort_values(by=['InputMass']).reset_index(
                    drop=True)
            x.close()
            return msg
        return read_file_w

    @pytest.fixture(scope="class")
    def check_file_contents(self, read_file):
        r"""Check that the contents of a file are correct.

        Args:
            fname (str): Full path to the file that should be checked.
            result (str): Contents of the file.

        """
        def check_file_contents_w(fname, result):
            ocont = read_file(fname)
            assert_frame_equal(ocont, result)
        return check_file_contents_w
    
    @pytest.fixture(scope="class")
    def check_file_size(self):
        r"""Check that file is the correct size.

        Args:
            fname (str): Full path to the file that should be checked.
            fsize (int): Size that the file should be in bytes.

        """
        def check_file_size_w(*args, **kwargs):
            pass
        return check_file_size_w



import pytest
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)


class TestTemptypeMetaschemaProperty(base_class):
    r"""Test class for TemptypeMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.AnyMetaschemaProperties'
    _cls = 'TemptypeMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return [(int(1), {'type': 'int'}),
                ([int(1), float(1)],
                 {'type': 'array',
                  'items': [{'type': 'int'}, {'type': 'float'}]}),
                ({'a': int(1), 'b': float(2)},
                 {'type': 'object',
                  'properties': {
                      'a': {'type': 'int'},
                      'b': {'type': 'float'}}})]

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return [(int(1), {'type': 'float'})]

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [({'type': 'int'}, {'type': 'int'})]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [({'type': 'int'}, {'type': 'float'})]



import pytest
from tests.metaschema.datatypes.test_PlyMetaschemaType import (
    TestPlyMetaschemaType as base_class)
from tests.metaschema.datatypes.test_PlyMetaschemaType import (
    TestPlyDict as base_class_dict)
import os
import copy
import shutil
import tempfile
from yggdrasil.metaschema.datatypes import ObjMetaschemaType


def test_create_schema():
    r"""Test create_schema."""
    with pytest.raises(RuntimeError):
        ObjMetaschemaType.create_schema(overwrite=False)
    temp = os.path.join(tempfile.gettempdir(), 'temp_schema')
    old_schema = ObjMetaschemaType.get_schema()
    try:
        shutil.move(ObjMetaschemaType._schema_file, temp)
        new_schema = ObjMetaschemaType.get_schema()
        assert(old_schema == new_schema)
    except BaseException:  # pragma: debug
        shutil.move(temp, ObjMetaschemaType._schema_file)
        raise
    shutil.move(temp, ObjMetaschemaType._schema_file)


class TestObjDict(base_class_dict):
    r"""Test for ObjDict class."""
    
    _mod = 'yggdrasil.metaschema.datatypes.ObjMetaschemaType'
    _cls = 'ObjDict'

    @pytest.fixture(scope="class")
    def simple_test(self, obj_test_value_simple):
        r"""dict: Simple test structure."""
        return obj_test_value_simple

    @pytest.fixture
    def instance_kwargs(self, obj_test_value):
        r"""Keyword arguments for a new instance of the tested class."""
        return obj_test_value

    @pytest.fixture(scope="class")
    def remove_hd_faces(self):
        r"""Remove higher dimension faces (>3 verts)."""
        def remove_hd_faces_w(obj):
            if 'faces' in obj:
                obj['faces'] = [f[:3] for f in obj['faces']]
            return obj
        return remove_hd_faces_w

    @pytest.fixture
    def objects_array_dict(self, instance):
        r"""Objects for testing transformation to/from dict of arrays."""
        orig = copy.deepcopy(instance)
        orig_edges = copy.deepcopy(instance)
        for v in orig_edges['texcoords']:
            v.pop('w', None)
        orig_nocolor = copy.deepcopy(instance)
        for v in orig_nocolor['vertices']:
            for k in ['red', 'green', 'blue']:
                v.pop(k, None)
        for v in orig_nocolor['params']:
            v.pop('w', None)
        for v in orig_nocolor['texcoords']:
            v.pop('v', None)
            v.pop('w', None)
        for surf in orig_nocolor['surfaces']:
            for v in surf['vertex_indices']:
                v.pop('texcoord_index', None)
                v.pop('normal_index', None)
        return [orig, orig_edges, orig_nocolor]
            
    @pytest.fixture
    def objects_trimesh(self, instance):
        r"""Objects for testing transformation to/from trimesh class."""
        if not ObjMetaschemaType.trimesh:
            pytest.skip("trimesh not available")
        orig = copy.deepcopy(instance)
        for k in ['material', 'params', 'points', 'surfaces', 'texcoords',
                  'curves', 'curve2Ds', 'lines', 'edges', 'normals']:
            orig.pop(k, None)
        for vlist in orig['faces']:
            for v in vlist:
                v.pop('texcoord_index', None)
                v.pop('normal_index', None)
        for v in orig['vertices']:
            for k in ['w']:  # , 'red', 'blue', 'green']:
                v.pop(k, None)
        return [orig]
            
    def test_vertex_normals(self, instance):
        r"""Test vertex_normals."""
        instance.vertex_normals

    def test_to_from_scene_incomplete(self, to_from_scene, simple_test):  # pragma: lpy
        r"""Test conversion to/from PlantGL scene with faces missing
        texcoords/normals on last element."""
        data = copy.deepcopy(simple_test)
        for f in data['faces']:
            for ff in f:
                ff.pop('texcoord_index', None)
                ff.pop('normal_index', None)
        to_from_scene(data)
        

class TestObjMetaschemaType(base_class):
    r"""Test class for ObjMetaschemaType class with float."""

    _mod = 'yggdrasil.metaschema.datatypes.ObjMetaschemaType'
    _cls = 'ObjMetaschemaType'

    @pytest.fixture(scope="class")
    def value(self, obj_test_value):
        r"""list: Test value."""
        return obj_test_value

    @pytest.fixture(scope="class")
    def valid_decoded(self, value, obj_test_value):
        r"""list: Objects that are valid under this type."""
        out = [value,
               ObjMetaschemaType.ObjDict(**obj_test_value),
               {'vertices': value['vertices'],
                'faces': [[{'vertex_index': 0},
                           {'vertex_index': 1},
                           {'vertex_index': 2}]]}]
        if ObjMetaschemaType.trimesh:
            out.append(ObjMetaschemaType.ObjDict(**obj_test_value).as_trimesh())
        return out
    
    @pytest.fixture(scope="class")
    def invalid_decoded(self, value):
        r"""list: Objects that are invalid under this type."""
        return [{'vertices': [{k: 0.0 for k in 'xyz'}],
                 'faces': [[{'vertex_index': 0},
                            {'vertex_index': 1},
                            {'vertex_index': 2}]]},
                {'vertices': [], 'faces': None,
                 'lines': [[None]],
                 'surfaces': [{'vertex_indices': [[]]}]},
                {'vertices': value['vertices'],
                 'texcoords': value['texcoords'],
                 'normals': value['normals'],
                 'faces': [[{'vertex_index': 0,
                             'texcoord_index': 100,
                             'normal_index': 100},
                            {'vertex_index': 1},
                            {'vertex_index': 2}]]},
                None]

    @pytest.fixture
    def nested_result(self, nested_approx):
        r"""Prepare value for comparison."""
        def nested_result_w(x):
            if ObjMetaschemaType.trimesh:
                if isinstance(x, ObjMetaschemaType.trimesh.base.Trimesh):
                    x = ObjMetaschemaType.ObjDict.from_trimesh(x)
            return nested_approx(x)
        return nested_result_w



import logging
import numpy as np
from yggdrasil.multitasking import _on_mpi, MPI, RLock, MPIRequestWrapper
from yggdrasil.communication import (
    CommBase, NoMessages)
logger = logging.getLogger(__name__)


class MPIClosedError(BaseException):
    r"""Exception to raise when the MPI connection has been closed."""


class MPIRequest(object):
    r"""Container for MPI request."""

    __slots__ = ['comm', 'address', 'direction', 'tag', 'size_req',
                 'req', 'size', '_data']

    def __init__(self, comm, direction, address, tag, **kwargs):
        self.comm = comm
        self.address = address
        self.direction = direction
        self.tag = tag
        self.size = np.zeros(1, dtype='i')
        self._data = None
        self.size_req = None
        self.req = self.make_request(**kwargs)

    @property
    def data(self):
        r"""object: Data returned by a request."""
        if (self.direction == 'recv') and isinstance(self._data, np.ndarray):
            self._data = self._data.tobytes()
        return self._data

    def make_request(self, payload=None):
        r"""Complete a request."""
        kwargs = dict(tag=self.tag)
        if self.direction == 'send':
            method = 'Isend'
            args = ([payload, MPI.CHAR], )
            kwargs['dest'] = self.address
            # Send size of message
            self.size[0] = len(payload)
            self.size_req = MPIRequestWrapper(
                self.comm.Isend([self.size, MPI.INT], **kwargs))
            out = MPIRequestWrapper(
                self.comm.Isend(*args, **kwargs))
        else:
            method = 'Irecv'
            args = ([self.size, MPI.INT], )
            kwargs['source'] = self.address
            self.size_req = MPIRequestWrapper(
                self.comm.Irecv(*args, **kwargs))
            out = None
        logger.debug("rank = %d, method = %s, args = %.100s, kwargs = %.100s",
                     self.comm.Get_rank(), method, args, kwargs)
        return out

    @property
    def complete(self):
        r"""bool: True if the request has been completed, False otherwise."""
        if self.direction == 'recv':
            if self.size_req.test()[0] and (self.req is None):
                self._data = np.zeros(self.size[0], dtype='c')
                self.req = MPIRequestWrapper(
                    self.comm.Irecv([self._data, MPI.CHAR],
                                    source=self.address,
                                    tag=self.tag))
        return (((self.req is not None)
                 and (self.req.test()[0] or self.req.canceled))
                or self.size_req.canceled)

    def cancel(self):
        r"""Cancel a request."""
        self.size_req.cancel()
        if self.req is not None:  # pragma: intermittent
            self.req.cancel()


class MPIMultiRequest(MPIRequest):
    r"""Container for MPI request for multiple partner comms."""

    __slots__ = ["remainder"]

    def __init__(self, *args, **kwargs):
        self.remainder = {}
        super(MPIMultiRequest, self).__init__(*args, **kwargs)

    def make_request(self, previous_requests=None):
        r"""Complete a request."""
        out = previous_requests
        if out is None:
            out = {}
        for x in self.address:
            if x not in out:
                out[x] = MPIRequest(self.comm, self.direction, x, self.tag[x])
        return out

    @property
    def complete(self):
        r"""bool: True if the request has been completed, False otherwise."""
        if not self.remainder:
            for k, v in self.req.items():
                if v.complete:
                    self._data = v.data
                    self.address = k
                    for k2, v2 in self.req.items():
                        if k2 != k:
                            self.remainder[k2] = v2
                    break
        return bool(self.remainder)

    def cancel(self):
        r"""Cancel a request."""
        for k, v in self.req.items():
            v.cancel()


class MPIComm(CommBase.CommBase):
    r"""Class for handling I/O via MPI communicators.

    Args:
        tag_start (int, optional): Tag that MPI messages should start with.
            Defaults to 0.
        tag_stride (int, optional): Amount that tag should be advanced after
            each message to avoid conflicts w/ other MPIComm communicators.
            Defaults to 1.
        partner_mpi_ranks (list, optional): Rank of MPI processes that partner
            models are running on. Defaults to None.

    Attributes:
        tag (int): Tag that should be used for the next MPI message.
        tag_stride (int): Amount that tag should be advanced after each
            each message to avoid conflicts w/ other MPIComm communicators.

    """

    _commtype = 'mpi'
    _schema_subtype_description = 'MPI communicator.'
    address_description = "The partner communicator ID(s)."
    _spacer_tags = 10
    _max_response = 10
    # Based on limit of 32bit int, this could be 2**30 - 1, but this is
    # too large for stack allocation in C so 2**20 will be used in case a
    # C implementation of the MPIComm is added in the future.
    _maxMsgSize = 2**20

    def __init__(self, *args, ranks=[], tag_start=0, tag_stride=1, **kwargs):
        assert(_on_mpi)
        if kwargs.get('partner_mpi_ranks', []):
            assert(kwargs.get('address', 'generate') in ['generate',
                                                         'address'])
            ranks = kwargs['partner_mpi_ranks']
        self._request_lock = RLock(task_method='thread')
        self.requests = []
        self.unused_tags = {}
        self.tags = {}
        self.ranks = ranks
        self.tag_start = tag_start
        self.tag_stride = tag_stride
        self.requires_disconnect = False
        self.last_request = None
        self.mpi_comm = MPI.COMM_WORLD
        self._is_open = False
        self.response_tags = []
        self.eof_recv = {}
        kwargs['no_suffix'] = True
        super(MPIComm, self).__init__(*args, **kwargs)
        self.eof_recv = {x: 0 for x in self.ranks}

    @classmethod
    def format_address(cls, ranks, tag_start, tag_stride):
        r"""Format an MPI address.

        Args:
            ranks (tuple): Ranks of the partner MPI processes.
            tag_start (int): Tag that the comm starts at.
            tag_stride (int): Tag that the comm advances by.

        Returns:
            str: Formatted address.

        """
        rank_str = '-'.join([str(x) for x in ranks])
        return f'{rank_str}_MPI_{tag_start}_MPI_{tag_stride}'

    @classmethod
    def parse_address(cls, address):
        r"""Parse an MPI address for information about the partner process
        ranks and how the tags should be iterated.

        Args:
            address (str): Address to parse.

        Returns:
            tuple: The ranks, starting tag, and tag stride contained in the
                address.

        """
        rank_str, tag_start, tag_stride = address.split('_MPI_')
        ranks = tuple([int(x) for x in rank_str.split('-')])
        return ranks, int(tag_start), int(tag_stride)

    @property
    def tag(self):
        r"""int: Tag for the next message."""
        return self.get_tag(max(self.ranks, key=self.get_tag))

    def next_rank(self):
        r"""Get the rank that should be used next."""
        if len(self.ranks) == 1:
            return self.ranks[0]
        elif self.direction == 'send':
            return min(self.ranks, key=self.get_tag)
        else:
            return self.ranks

    def get_tag(self, rank=None):
        r"""Get the next tag for a rank.

        Args:
            rank (int): Rank to get tag for.

        Returns:
            int: Tag that should be used next for the rank.

        """
        if rank is None:
            rank = self.next_rank()
        if isinstance(rank, (list, tuple)):
            return {x: self.get_tag(x) for x in rank}
        if self.unused_tags.get(rank, []):
            return self.unused_tags[rank][0]
        return self.tags[rank]

    def advance_tag(self, request):
        r"""Advance to the next tag.

        Args:
            request (MPIRequest, MPIMultiRequest): Request advancing the tag.

        """
        if isinstance(request, MPIMultiRequest):
            for v in request.req.values():
                self.advance_tag(v)
            return
        if request.tag in self.unused_tags.get(request.address, []):
            self.unused_tags[request.address].remove(request.tag)
            return
        self.tags[request.address] = max(self.tags[request.address],
                                         request.tag + self.tag_stride)

    def cache_tag(self, request):
        r"""Store a tag for an uncompleted request.

        Args:
            request (MPIRequest, MPIMultiRequest): Request to cache.

        """
        if isinstance(request, MPIMultiRequest):
            for v in request.req.values():
                self.cache_tag(v)
            return
        self.unused_tags.setdefault(request.address, [])
        self.unused_tags[request.address].append(request.tag)

    def bind(self):
        r"""Bind to random queue if address is generate."""
        assert(isinstance(self.address, str))
        if self.address in ['generate', 'address']:
            assert(self.ranks)
            self.address = self.format_address(self.ranks, self.tag_start,
                                               self.tag_stride)
        else:
            self.ranks, self.tag_start, self.tag_stride = self.parse_address(
                self.address)
        if not self.tags:
            self.tags = {x: self.tag_start for x in self.ranks}
        super(MPIComm, self).bind()

    @property
    def model_env(self):
        r"""dict: Mapping between model name and opposite comm
        environment variables that need to be provided to the model."""
        return {}
    
    @property
    def opp_address(self):
        r"""str: Address for opposite comm."""
        return self.format_address([self.mpi_comm.Get_rank()],
                                   self.tag_start, self.tag_stride)
        
    @property
    def get_response_comm_kwargs(self):
        r"""dict: Keyword arguments to use for a response comm."""
        out = super(MPIComm, self).get_response_comm_kwargs
        # tag = self.tag_start + len(self.response_tags) + 2
        # tag_stride = self.tag_stride
        # if (tag - self.tag_start) >= self._max_response:
        #     raise Exception("Starting tag for next response comm "
        #                     "exceeds the maximum and may conflict with "
        #                     "other messages this comm will send.")
        tag = self.get_tag()
        tag_stride = 0
        assert(tag not in self.response_tags)
        self.response_tags.append(tag)
        out['address'] = self.format_address(
            [self.next_rank()], tag, tag_stride)
        return out

    @property
    def create_work_comm_kwargs(self):
        r"""dict: Keyword arguments for a new work comm."""
        out = super(MPIComm, self).create_work_comm_kwargs
        out['address'] = self.format_address(
            [self.next_rank()], self.get_tag(), 0)
        return out
    
    def open(self):
        r"""Open the queue."""
        super(MPIComm, self).open()
        if not self.is_open:
            self._is_open = True
            assert(self.mpi_comm.Get_rank() not in self.ranks)
            
    def _close(self, linger=False):
        r"""Close the queue."""
        # Disconnect will only be required if a subset of processes is used,
        # but that is not currently supported. This should be uncommented if
        # support for dynamic processes management is added.
        # if self.requires_disconnect and self.is_open:
        #     self.mpi_comm.Disconnect()
        with self._request_lock:
            self.mpi_comm = None
            self.cancel_requests()
        super(MPIComm, self)._close(linger=linger)

    def cancel_requests(self):
        r"""Cancel requests that have not yet been completed."""
        with self._request_lock:
            complete_requests = []
            for x in self.requests:
                if x.complete:
                    complete_requests.append(x)
                else:
                    self.cache_tag(x)
                    x.cancel()
            # Cancel uncompleted partial request for multi-receive?
            self.requests = complete_requests

    @property
    def is_open(self):
        r"""bool: True if the queue is not None."""
        return (self.mpi_comm is not None) and self._is_open
        
    def confirm_send(self, noblock=False):
        r"""Confirm that sent message was received."""
        if noblock:
            return True
        return (self.n_msg_send == 0)

    def confirm_recv(self, noblock=False):
        r"""Confirm that message was received."""
        if noblock:
            return True
        return (self.n_msg_recv == 0)

    @property
    def n_msg_send(self):
        r"""int: Number of messages in the queue to send."""
        if self.is_open and self.requests:
            return sum([(not x.complete) for x in self.requests])
        else:
            return 0
        
    @property
    def n_msg_recv(self):
        r"""int: Number of messages in the queue to recv."""
        if self.is_open:
            try:
                self.add_request(on_empty=True)
            except MPIClosedError:  # pragma: intermittent
                return 0
        if self.is_open and self.requests:
            return sum([x.complete for x in self.requests])
        else:
            return 0

    def add_request(self, on_empty=False, **kwargs):
        r"""Add a request to the queue."""
        with self._request_lock:
            if on_empty and self.requests:
                return
            cls = MPIRequest
            address = self.next_rank()
            tag = self.get_tag(address)
            if isinstance(address, (list, tuple)):
                if self.last_request:
                    kwargs['previous_requests'] = self.last_request.remainder
                cls = MPIMultiRequest
            if self.mpi_comm is None:  # pragma: intermittent
                raise MPIClosedError
            args = (self.mpi_comm, self.direction, address, tag)
            req = cls(*args, **kwargs)
            self.requests.append(req)
            self.advance_tag(req)

    def send_message(self, msg, **kwargs):
        r"""Send a message encapsulated in a CommMessage object.

        Args:
            msg (CommMessage): Message to be sent.
            **kwargs: Additional keyword arguments are passed to _safe_send.

        Returns:
            bool: Success or failure of send.
        
        """
        if (msg.flag == CommBase.FLAG_EOF) and (msg.args != 'DONT_RECURSE'):
            for _ in range(len(self.ranks) - 1):
                msg.add_message(msg=msg.msg, length=msg.length, flag=msg.flag,
                                args='DONT_RECURSE', header=msg.header)
        return super(MPIComm, self).send_message(msg, **kwargs)

    def recv_message(self, *args, **kwargs):
        r"""Receive a message.

        Args:
            *args: Arguments are passed to the forked comm's recv_message method.
            **kwargs: Keyword arguments are passed to the forked comm's recv_message
                method.

        Returns:
            CommMessage: Received message.

        """
        out = super(MPIComm, self).recv_message(*args, **kwargs)
        self.sleep()
        if out.flag == CommBase.FLAG_EOF:
            self.eof_recv[self.last_request.address] = 1
            if not all(self.eof_recv.values()):
                out = self.recv_message(*args, **kwargs)
        return out
        
    def _send(self, payload):
        r"""Send a message.

        Args:
            payload (str): Message to send.

        Returns:
            bool: Success or failure of sending the message.

        """
        self.add_request(payload=payload)
        return True
        
    def _recv(self):
        r"""Receive a message from the MPI communicator.

        Returns:
            tuple (bool, str): The success or failure of receiving a message
                and the message received.

        """
        with self._request_lock:
            self.add_request(on_empty=True)
            if not self.requests[0].complete:
                raise NoMessages("No messages in communicator.")
            self.last_request = self.requests.pop(0)
            return (True, self.last_request.data)

    def purge(self):
        r"""Purge all messages from the comm."""
        super(MPIComm, self).purge()
        with self._request_lock:
            self.cancel_requests()
            while self.n_msg_recv > 0:  # pragma: intermittent
                self._recv()



import copy
import pytest
from yggdrasil.communication import CommBase, new_comm, get_comm
from tests.communication.test_CommBase import TestComm as base_class


@pytest.mark.suite('mpi')
@pytest.mark.mpi(min_size=2)
class TestMPIComm(base_class):
    r"""Test class for MPIComm."""
    # Rank 0 sends, rank 1+ receives

    test_error_send = None
    test_error_recv = None
    test_invalid_direction = None
    test_work_comm = None

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "mpi"

    @pytest.fixture(scope="class")
    def root_direction(self):
        r"""str: Direction of communicator on the root process."""
        return 'send'

    @pytest.fixture(scope="class")
    def rank_direction(self, mpi_rank, root_direction):
        r"""str: MPI rank of the direction this processes is handling
        messages."""
        if mpi_rank == 0:
            return root_direction
        elif root_direction == 'recv':
            return 'send'
        else:
            return 'recv'

    @pytest.fixture
    def get_next_tag(self, adv_global_mpi_tag, initial_tag):
        def wrapped_get_next_tag():
            return adv_global_mpi_tag()
        return wrapped_get_next_tag

    @pytest.fixture(scope="class")
    def sync(self, sync_mpi_exchange):
        def wrapped_sync(local_comm, **kwargs):
            return sync_mpi_exchange(local_comm.tag, **kwargs)
        return wrapped_sync

    @pytest.fixture
    def initial_tag(self, adv_global_mpi_tag, mpi_rank):
        return adv_global_mpi_tag()

    @pytest.fixture
    def root_kwargs(self, commtype, use_async, testing_options,
                    root_direction, get_next_tag, mpi_rank, mpi_size):
        r"""dict: Keyword arguments for the communicator on the rank == 0
        process."""
        if mpi_rank == 0:
            out = dict(commtype=commtype, reverse_names=True,
                       direction=root_direction, use_async=use_async)
            out.update(testing_options['kwargs'],
                       tag_start=get_next_tag(),
                       partner_mpi_ranks=list(range(1, mpi_size)))
        else:
            out = {}
        return out

    @pytest.fixture
    def rank_kwargs(self, mpi_comm, mpi_rank, initial_tag,
                    get_next_tag, commtype):
        r"""dict: Keyword arguments for the communicator on the rank > 0
        process(es)."""
        if mpi_rank > 0:
            out = mpi_comm.recv(source=0, tag=initial_tag)
            out.update(tag_start=get_next_tag(),
                       commtype=commtype)
        else:
            out = {}
        return out

    # Disable global comms for MPI
    @pytest.fixture
    def global_send_comm(self, send_comm):
        r"""Communicator for sending messages."""
        yield send_comm

    @pytest.fixture
    def global_recv_comm(self, recv_comm):
        r"""Communicator for receiving messages."""
        yield recv_comm

    @pytest.fixture
    def global_comm(self, global_recv_comm, global_send_comm):
        r"""Global communicator."""
        if global_recv_comm is None:
            return global_send_comm
        return global_recv_comm
    
    @pytest.fixture(scope="class")
    def sleep_after_connect(self):
        r"""Indicates if sleep should occur after comm creation."""
        return False
    
    @pytest.fixture(scope="class")
    def alt_sleep_after_connect(self):
        r"""Indicates if sleep should occur after comm creation."""
        return False

    # TODO: Adjust nprev_comm & nprev_fd?
    
    @pytest.fixture
    def root_comm(self, name, mpi_comm, mpi_rank, root_kwargs,
                  initial_tag, close_comm):
        r"""MPI communicator for rank == 0 process."""
        if mpi_rank == 0:
            x = new_comm(name, **root_kwargs)
            rank_kws = x.opp_comm_kwargs()
            for i in x.ranks:
                mpi_comm.send(rank_kws, dest=i, tag=initial_tag)
            yield x
            close_comm(x)
        else:
            yield None
        
    @pytest.fixture
    def rank_comm(self, name, mpi_rank, rank_kwargs, close_comm):
        r"""MPI communicator for rank > 0 processes."""
        if mpi_rank > 0:
            x = get_comm(name, **rank_kwargs)
            yield x
            close_comm(x)
        else:
            yield None

    @pytest.fixture
    def local_comm(self, mpi_rank, root_comm, rank_comm, sync,
                   alt_sleep_after_connect, use_async):
        r"""Communicator for the current MPI process."""
        if mpi_rank == 0:
            x = root_comm
        else:
            x = rank_comm
        if alt_sleep_after_connect:  # pragma: testing
            x.sleep()
        assert(x.is_open)
        if x.direction == 'recv':
            x.drain_server_signon_messages()
        sync(x)
        yield x
        # Even up send/recv calls since the same comm will be used for
        # subsequent tests
        if use_async and (x.direction == 'recv'):
            # Don't keep receiving or there will never be a balence
            # between receive requests and sent messages
            x._close_backlog(wait=True)
        all_tag = sync(x, get_tags=True)
        if x.direction == 'send':
            for _ in range(max(all_tag) - all_tag[mpi_rank]):
                x.tags[x.ranks[0]] += 1
            if use_async:
                x.wait_for_confirm(timeout=60.0)
        sync(x, check_equal=True)
    
    @pytest.fixture
    def recv_comm(self, rank_direction, local_comm):
        r"""Receiving communicator."""
        if rank_direction == 'recv':
            return local_comm

    @pytest.fixture
    def send_comm(self, rank_direction, local_comm):
        r"""Sending communicator."""
        if rank_direction == 'send':
            return local_comm

    def test_send_recv_after_close(self, local_comm, testing_options):
        r"""Test that opening twice dosn't cause errors and that send/recv
        after close returns false."""
        local_comm.open()
        local_comm.close()
        assert(local_comm.is_closed)
        if local_comm.direction == 'send':
            flag = local_comm.send(testing_options['msg'])
        elif local_comm.direction == 'recv':
            flag, msg_recv = local_comm.recv()
        assert(not flag)

    def test_send_after_close(self, use_async, local_comm, testing_options,
                              sync):
        r"""Sending a message after the receive comm has closed."""
        if use_async:
            pytest.skip("skip for async")
        if local_comm.direction == 'recv':
            local_comm.close()
            for i in local_comm.tags.keys():
                local_comm.tags[i] += 1
        sync(local_comm)
        if local_comm.direction == 'send':
            flag = local_comm.send(testing_options['msg'])
            assert(flag)

    # def test_work_comm(self, local_comm, testing_options, uuid, timeout):
    #     r"""Test creating/removing a work comm."""
    #     wc_send = local_comm.create_work_comm()
    #     with pytest.raises(KeyError):
    #         local_comm.add_work_comm(wc_send)
    #     # Create recv instance in way that tests new_comm
    #     header_recv = dict(id=uuid + '1', address=wc_send.address)
    #     recv_kwargs = local_comm.get_work_comm_kwargs
    #     recv_kwargs.pop('async_recv_kwargs', None)
    #     recv_kwargs['work_comm_name'] = 'test_worker_%s' % header_recv['id']
    #     recv_kwargs['commtype'] = wc_send._commtype
    #     if isinstance(wc_send.opp_address, str):
    #         os.environ[recv_kwargs['work_comm_name']] = wc_send.opp_address
    #     else:
    #         recv_kwargs['address'] = wc_send.opp_address
    #     wc_recv = local_comm.create_work_comm(**recv_kwargs)
    #     # wc_recv = local_comm.get_work_comm(header_recv)
    #     flag = wc_send.send(testing_options['msg'])
    #     assert(flag)
    #     flag, msg_recv = wc_recv.recv(timeout)
    #     assert(flag)
    #     assert(msg_recv == testing_options['msg'])
    #     # Assert errors on second attempt
    #     # with pytest.raises(RuntimeError):
    #     #     wc_send.send(testing_options['msg'])
    #     with pytest.raises(RuntimeError):
    #         wc_recv.recv()
    #     local_comm.remove_work_comm(wc_send.uuid)
    #     local_comm.remove_work_comm(wc_recv.uuid)
    #     local_comm.remove_work_comm(wc_recv.uuid)
    #     # Create work comm that should be cleaned up on teardown
    #     local_comm.create_work_comm()

    @pytest.fixture(scope="class")
    def do_send_recv(self, wait_on_function, testing_options, map_sent2recv,
                     n_msg_expected, sync, timeout, nested_approx, logger):
        r"""Factory for method to perform send/recv checks for comms."""

        def do_send(send_comm, send_params):
            assert(send_comm.n_msg_send == send_params.get('n_init', 0))
            send_params.setdefault('count', 1)
            if 'eof' not in send_params.get('method', 'send'):
                send_params['count'] *= len(send_comm.ranks)
            sync(send_comm)
            logger.debug(f"sending {send_params.get('count', 1)} "
                         f"copies of {send_params['message']!s:.100}")
            for _ in range(send_params.get('count', 1)):
                flag = getattr(send_comm, send_params.get('method', 'send'))(
                    *send_params['args'], **send_params.get('kwargs', {}))
                assert(flag == send_params.get('flag', True))
            if not send_params.get('skip_wait', False):
                wait_on_function(
                    lambda: send_comm.is_closed or (send_comm.n_msg_send == 0))
            send_comm.printStatus(level='debug')
            if 'eof' not in send_params.get('method', 'send'):
                send_comm.wait_for_confirm(timeout=timeout)
                assert(send_comm.is_confirmed)
                send_comm.confirm(noblock=True)
            assert(send_comm.n_msg_send == 0)
            sync(send_comm)
            
        def do_recv(recv_comm, recv_params):
            if (((not recv_comm.is_eof(recv_params['message']))
                 and (recv_params['message'] != b''))):
                recv_params['count'] *= len(recv_comm.ranks)
            assert(recv_comm.n_msg_recv == recv_params.get('n_init', 0))
            sync(recv_comm)
            logger.debug(f"expecting {recv_params.get('count', 1)} "
                         f"copies of {recv_params['message']!s:.100}")
            for _ in range(recv_params.get('count', 1)):
                if not recv_params.get('skip_wait', False):
                    wait_on_function(
                        lambda: (recv_comm.is_closed
                                 or (recv_comm.n_msg_recv > 0)))
                flag, msg = getattr(
                    recv_comm, recv_params.get('method', 'recv'))(
                        **recv_params.get('kwargs', {'timeout': 0.1}))
                assert(flag == recv_params.get('flag', True))
                assert(msg == nested_approx(recv_params['message']))
            recv_comm.printStatus()
            if not recv_comm.is_eof(recv_params['message']):
                recv_comm.wait_for_confirm(timeout=timeout)
                assert(recv_comm.is_confirmed)
                recv_comm.confirm(noblock=True)
            assert(recv_comm.n_msg_recv == 0)
            sync(recv_comm)

        def wrapped(send_comm, recv_comm, message=None,
                    send_params=None, recv_params=None):
            if send_comm is None:
                local_comm = recv_comm
            else:
                local_comm = send_comm
            if send_params is None:
                send_params = {}
            if recv_params is None:
                recv_params = {}
            local_comm.printStatus(level='debug')
            if send_params.get('method', 'send') == 'send_eof':
                message = local_comm.eof_msg
                send_params['args'] = tuple([])
            else:
                if message is None:
                    message = testing_options['msg']
                send_params['args'] = (copy.deepcopy(message),)
            send_params['message'] = message
            if (((not recv_params.get('skip_wait', False))
                 and ('eof' not in send_params.get('method', 'send')))):
                recv_params.setdefault('count', n_msg_expected)
            if 'message' not in recv_params:
                if 'eof' in send_params.get('method', 'send'):
                    recv_params['message'] = message
                else:
                    recv_params['message'] = map_sent2recv(message)
            if local_comm.direction == "send":
                do_send(local_comm, send_params)
            else:
                do_recv(local_comm, recv_params)
        return wrapped

    def test_cleanup_comms(self, local_comm):
        r"""Test cleanup_comms for comm class."""
        local_comm.cleanup_comms()
        assert(len(local_comm.comm_registry()) == 0)

    def test_drain_messages(self, local_comm, timeout):
        r"""Test waiting for messages to drain."""
        local_comm.drain_messages(timeout=timeout)
        assert(getattr(local_comm,
                       'n_msg_%s_drain' % local_comm.direction) == 0)
        with pytest.raises(ValueError):
            local_comm.drain_messages(variable='n_msg_invalid')

    def test_recv_nomsg(self, local_comm, polling_interval):
        r"""Test recieve when there is no waiting message."""
        if local_comm.direction == 'recv':
            flag, msg_recv = local_comm.recv(timeout=polling_interval)
            assert(flag)
            assert(not msg_recv)
        else:
            local_comm.sleep()

    @pytest.fixture
    def filtered_comms(self, local_comm, msg_filter_send, msg_filter_recv):
        r"""Add filters to the send and receive communicators."""
        from yggdrasil.communication.filters.StatementFilter import (
            StatementFilter)
        from yggdrasil.communication.filters.FunctionFilter import (
            FunctionFilter)
        if local_comm.direction == 'send':
            # Statement filter on send comm
            if isinstance(msg_filter_send, (str, bytes)):
                statement = '%x% != ' + repr(msg_filter_send)
            else:
                statement = 'repr(%x%) != r"""' + repr(msg_filter_send) + '"""'
            local_comm.filter = StatementFilter(statement=statement)
        else:
            # Function filter on recv comm

            def fcond(x):
                try:
                    assert(x == msg_filter_recv)
                    return False
                except BaseException:
                    return True
            local_comm.filter = FunctionFilter(function=fcond)
        yield
        local_comm.filter = None

    def test_send_recv_raw(self, local_comm, testing_options):
        r"""Test send/recv of a small message."""
        def dummy(msg):
            print(msg)
            msg.msg = b''
            msg.args = b''
            msg.length = 0
            return msg

        if local_comm.direction == 'send':
            for _ in range(len(local_comm.ranks)):
                assert(local_comm.send(testing_options['msg']))
        else:
            for _ in range(len(local_comm.ranks)):
                msg = local_comm.recv(
                    timeout=60.0, skip_deserialization=True,
                    return_message_object=True,
                    after_finalize_message=[dummy])
                assert(msg.finalized)
                assert(local_comm.finalize_message(msg) == msg)
                msg.finalized = False
                assert(local_comm.is_empty_recv(msg.args))
                msg = local_comm.finalize_message(msg)
                assert(msg.flag == CommBase.FLAG_EMPTY)

    def test_purge(self, use_async, local_comm, testing_options,
                   wait_on_function, n_msg_expected, sync):
        r"""Test purging messages from the comm."""
        assert(local_comm.n_msg == 0)
        if local_comm.direction == 'send':
            if local_comm.is_async:
                assert(local_comm.n_msg_direct == 0)
            sync(local_comm)
            for _ in range(len(local_comm.ranks)):
                flag = local_comm.send(testing_options['msg'])
            assert(flag)
            # local_comm.purge()
            # assert(local_comm.n_msg == 0)
            wait_on_function(lambda: local_comm.n_msg == 0)
            if use_async:
                local_comm._wrapped.wait_for_confirm()
            # sync(local_comm)
        else:
            if local_comm.is_async:
                assert(local_comm.n_msg_direct == 0)
            sync(local_comm)
            wait_on_function(lambda: local_comm.n_msg == n_msg_expected)
            assert(local_comm.n_msg > 0)
            local_comm.purge()
            assert(local_comm.n_msg == 0)
            # Purge recv while closed
            local_comm.close()
            local_comm.purge()
            # sync(local_comm)


@pytest.mark.mpi(min_size=3)
class TestMPICommMultipleRecv(TestMPIComm):
    r"""Test class for MPIComm with comm receiving from multiple processes."""

    @pytest.fixture(scope="class")
    def root_direction(self):
        r"""str: Direction of communicator on the root process."""
        return 'recv'



"""Constants used by yggdrasil."""
import numpy as np
from collections import OrderedDict
# No other yggdrasil modules should be import here
# TODO: Move platform constants into this module?
from yggdrasil import platform


# Type related constants
NUMPY_NUMERIC_TYPES = [
    'int',
    'uint',
    'float',
    'complex',
]
NUMPY_STRING_TYPES = [
    'bytes',
    'unicode',
    'str',
]
NUMPY_TYPES = NUMPY_NUMERIC_TYPES + NUMPY_STRING_TYPES
FLEXIBLE_TYPES = [
    'string',
    'bytes',
    'unicode',
]
PYTHON_SCALARS = OrderedDict([
    ('float', [float]),
    ('int', [int]),
    ('uint', []),
    ('complex', [complex]),
    ('bytes', [bytes]),
    ('unicode', [str]),
])
VALID_TYPES = OrderedDict([(k, k) for k in NUMPY_NUMERIC_TYPES])
VALID_TYPES.update([
    ('bytes', 'bytes'),
    ('unicode', 'str'),
])
NUMPY_PRECISIONS = {
    'float': [16, 32, 64],
    'int': [8, 16, 32, 64],
    'uint': [8, 16, 32, 64],
    'complex': [64, 128],
}
if not platform._is_win:
    # Not available on windows
    NUMPY_PRECISIONS['float'].append(128)
    NUMPY_PRECISIONS['complex'].append(256)
for T, T_NP in VALID_TYPES.items():
    PYTHON_SCALARS[T].append(np.dtype(T_NP).type)
    if T in NUMPY_PRECISIONS:
        PYTHON_SCALARS[T] += [np.dtype(T_NP + str(P)).type
                              for P in NUMPY_PRECISIONS[T]]
PYTHON_SCALARS['int'].append(np.signedinteger)
PYTHON_SCALARS['uint'].append(np.unsignedinteger)
ALL_PYTHON_SCALARS = []
for k, v in PYTHON_SCALARS.items():
    PYTHON_SCALARS[k] = tuple(v)
    ALL_PYTHON_SCALARS += list(v)
ALL_PYTHON_SCALARS = tuple(set(ALL_PYTHON_SCALARS))
ALL_PYTHON_ARRAYS = (np.ndarray,)


# Serialization constants
FMT_CHAR = b'%'
YGG_MSG_HEAD = b'YGG_MSG_HEAD'
DEFAULT_COMMENT = b'# '
DEFAULT_DELIMITER = b'\t'
DEFAULT_NEWLINE = b'\n'
FMT_CHAR_STR = FMT_CHAR.decode("utf-8")
DEFAULT_COMMENT_STR = DEFAULT_COMMENT.decode("utf-8")
DEFAULT_DELIMITER_STR = DEFAULT_DELIMITER.decode("utf-8")
DEFAULT_NEWLINE_STR = DEFAULT_NEWLINE.decode("utf-8")


# Communication constants
YGG_MSG_EOF = b'EOF!!!'
YGG_MSG_BUF = 1024 * 2
YGG_CLIENT_INI = b'YGG_BEGIN_CLIENT'
YGG_CLIENT_EOF = b'YGG_END_CLIENT'


# ======================================================
# Do not edit this file past this point as the following
# is generated by yggdrasil.schema.update_constants
# ======================================================


# Component registry
COMPONENT_REGISTRY = {
    'comm': {
        'base': 'CommBase',
        'default': 'default',
        'key': 'commtype',
        'module': 'yggdrasil.communication',
        'subtypes': {
            'buffer': 'BufferComm',
            'default': 'DefaultComm',
            'ipc': 'IPCComm',
            'mpi': 'MPIComm',
            'rest': 'RESTComm',
            'rmq': 'RMQComm',
            'rmq_async': 'RMQAsyncComm',
            'value': 'ValueComm',
            'zmq': 'ZMQComm',
        },
    },
    'connection': {
        'base': 'ConnectionDriver',
        'default': None,
        'key': 'connection_type',
        'module': 'yggdrasil.drivers',
        'subtypes': {
            'connection': 'ConnectionDriver',
            'file_input': 'FileInputDriver',
            'file_output': 'FileOutputDriver',
            'input': 'InputDriver',
            'output': 'OutputDriver',
            'rpc_request': 'RPCRequestDriver',
            'rpc_response': 'RPCResponseDriver',
        },
    },
    'file': {
        'base': 'FileComm',
        'default': 'binary',
        'key': 'filetype',
        'module': 'yggdrasil.communication',
        'subtypes': {
            'ascii': 'AsciiFileComm',
            'binary': 'FileComm',
            'json': 'JSONFileComm',
            'map': 'AsciiMapComm',
            'mat': 'MatFileComm',
            'netcdf': 'NetCDFFileComm',
            'obj': 'ObjFileComm',
            'pandas': 'PandasFileComm',
            'pickle': 'PickleFileComm',
            'ply': 'PlyFileComm',
            'table': 'AsciiTableComm',
            'wofost': 'WOFOSTParamFileComm',
            'yaml': 'YAMLFileComm',
        },
    },
    'filter': {
        'base': 'FilterBase',
        'default': None,
        'key': 'filtertype',
        'module': 'yggdrasil.communication.filters',
        'subtypes': {
            'direct': 'DirectFilter',
            'function': 'FunctionFilter',
            'statement': 'StatementFilter',
        },
    },
    'model': {
        'base': 'ModelDriver',
        'default': 'executable',
        'key': 'language',
        'module': 'yggdrasil.drivers',
        'subtypes': {
            'R': 'RModelDriver',
            'c': 'CModelDriver',
            'c++': 'CPPModelDriver',
            'cmake': 'CMakeModelDriver',
            'cpp': 'CPPModelDriver',
            'cxx': 'CPPModelDriver',
            'dummy': 'DummyModelDriver',
            'executable': 'ExecutableModelDriver',
            'fortran': 'FortranModelDriver',
            'lpy': 'LPyModelDriver',
            'make': 'MakeModelDriver',
            'matlab': 'MatlabModelDriver',
            'mpi': 'MPIPartnerModel',
            'osr': 'OSRModelDriver',
            'python': 'PythonModelDriver',
            'r': 'RModelDriver',
            'sbml': 'SBMLModelDriver',
            'timesync': 'TimeSyncModelDriver',
        },
    },
    'serializer': {
        'base': 'SerializeBase',
        'default': 'default',
        'key': 'seritype',
        'module': 'yggdrasil.serialize',
        'subtypes': {
            'default': 'DefaultSerialize',
            'direct': 'DirectSerialize',
            'functional': 'FunctionalSerialize',
            'json': 'JSONSerialize',
            'map': 'AsciiMapSerialize',
            'mat': 'MatSerialize',
            'obj': 'ObjSerialize',
            'pandas': 'PandasSerialize',
            'pickle': 'PickleSerialize',
            'ply': 'PlySerialize',
            'table': 'AsciiTableSerialize',
            'wofost': 'WOFOSTParamSerialize',
            'yaml': 'YAMLSerialize',
        },
    },
    'transform': {
        'base': 'TransformBase',
        'default': None,
        'key': 'transformtype',
        'module': 'yggdrasil.communication.transforms',
        'subtypes': {
            'array': 'ArrayTransform',
            'direct': 'DirectTransform',
            'filter': 'FilterTransform',
            'function': 'FunctionTransform',
            'iterate': 'IterateTransform',
            'map_fields': 'MapFieldsTransform',
            'pandas': 'PandasTransform',
            'select_fields': 'SelectFieldsTransform',
            'statement': 'StatementTransform',
        },
    },
}

# Language driver constants
LANG2EXT = {
    'R': '.R',
    'c': '.c',
    'c++': '.cpp',
    'cpp': '.cpp',
    'cxx': '.cpp',
    'executable': '.exe',
    'fortran': '.f90',
    'lpy': '.lpy',
    'matlab': '.m',
    'osr': '.xml',
    'python': '.py',
    'r': '.R',
    'sbml': '.xml',
    'yaml': '.yml',
}
EXT2LANG = {v: k for k, v in LANG2EXT.items()}
LANGUAGES = {
    'compiled': [
        'c', 'c++', 'fortran'],
    'interpreted': [
        'R', 'matlab', 'python'],
    'build': [
        'cmake', 'make'],
    'dsl': [
        'lpy', 'osr', 'sbml'],
    'other': [
        'dummy', 'executable', 'mpi', 'timesync'],
}
LANGUAGES['all'] = (
    LANGUAGES['compiled']
    + LANGUAGES['interpreted']
    + LANGUAGES['build']
    + LANGUAGES['dsl']
    + LANGUAGES['other'])
LANGUAGES_WITH_ALIASES = {
    'compiled': [
        'c', 'c++', 'cpp', 'cxx', 'fortran'],
    'interpreted': [
        'R', 'matlab', 'python', 'r'],
    'build': [
        'cmake', 'make'],
    'dsl': [
        'lpy', 'osr', 'sbml'],
    'other': [
        'dummy', 'executable', 'mpi', 'timesync'],
}
LANGUAGES_WITH_ALIASES['all'] = (
    LANGUAGES_WITH_ALIASES['compiled']
    + LANGUAGES_WITH_ALIASES['interpreted']
    + LANGUAGES_WITH_ALIASES['build']
    + LANGUAGES_WITH_ALIASES['dsl']
    + LANGUAGES_WITH_ALIASES['other'])
ALIASED_LANGUAGES = {
    'R': [
        'R', 'r'],
    'c++': [
        'c++', 'cpp', 'cxx'],
}
COMPILER_ENV_VARS = {
    'c': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'c++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'fortran': {
        'exec': 'FC',
        'flags': 'FFLAGS',
    },
}
COMPILATION_TOOL_VARS = {
    'LIB': {
        'exec': None,
        'flags': None,
    },
    'LINK': {
        'exec': 'LINK',
        'flags': 'LDFLAGS',
    },
    'ar': {
        'exec': 'AR',
        'flags': None,
    },
    'cl': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'cl++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'clang': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'clang++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'cmake': {
        'exec': None,
        'flags': None,
    },
    'g++': {
        'exec': 'CXX',
        'flags': 'CXXFLAGS',
    },
    'gcc': {
        'exec': 'CC',
        'flags': 'CFLAGS',
    },
    'gfortran': {
        'exec': 'FC',
        'flags': 'FFLAGS',
    },
    'ld': {
        'exec': 'LD',
        'flags': 'LDFLAGS',
    },
    'libtool': {
        'exec': 'LIBTOOL',
        'flags': None,
    },
    'make': {
        'exec': None,
        'flags': None,
    },
    'nmake': {
        'exec': None,
        'flags': None,
    },
}
LANGUAGE_PROPERTIES = {
    'R': {
        'executable_type': 'interpreted',
        'full_language': True,
        'is_typed': False,
    },
    'c': {
        'executable_type': 'compiled',
        'full_language': True,
        'is_typed': True,
    },
    'c++': {
        'executable_type': 'compiled',
        'full_language': True,
        'is_typed': True,
    },
    'cmake': {
        'executable_type': 'build',
        'full_language': False,
        'is_typed': False,
    },
    'dummy': {
        'executable_type': 'other',
        'full_language': False,
        'is_typed': False,
    },
    'executable': {
        'executable_type': 'other',
        'full_language': False,
        'is_typed': False,
    },
    'fortran': {
        'executable_type': 'compiled',
        'full_language': True,
        'is_typed': True,
    },
    'lpy': {
        'executable_type': 'dsl',
        'full_language': False,
        'is_typed': False,
    },
    'make': {
        'executable_type': 'build',
        'full_language': False,
        'is_typed': False,
    },
    'matlab': {
        'executable_type': 'interpreted',
        'full_language': True,
        'is_typed': False,
    },
    'mpi': {
        'executable_type': 'other',
        'full_language': False,
        'is_typed': False,
    },
    'osr': {
        'executable_type': 'dsl',
        'full_language': False,
        'is_typed': False,
    },
    'python': {
        'executable_type': 'interpreted',
        'full_language': True,
        'is_typed': False,
    },
    'sbml': {
        'executable_type': 'dsl',
        'full_language': False,
        'is_typed': False,
    },
    'timesync': {
        'executable_type': 'other',
        'full_language': False,
        'is_typed': False,
    },
}



import pytest
from tests.serialize.test_AsciiTableSerialize import (
    TestAsciiTableSerialize as base_class)


@pytest.mark.usefixtures("pandas_equality_patch")
class TestPandasSerialize(base_class):
    r"""Test class for TestPandasSerialize class."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "pandas"

    @pytest.fixture(scope="class", autouse=True,
                    params=[{}, {'no_header': True},
                            {'table_string_type': 'bytes'}])
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return request.param

    def test_apply_field_names_errors(self, instance, testing_options):
        r"""Test errors raised by apply_field_names."""
        with pytest.raises(RuntimeError):
            instance.apply_field_names(testing_options['objects'][0],
                                       field_names=['x', 'y'])
        names = testing_options['objects'][0].columns.tolist()
        names[0] = 'invalid'
        with pytest.raises(RuntimeError):
            instance.apply_field_names(testing_options['objects'][0],
                                       field_names=names)

    def test_func_serialize_errors(self, instance):
        r"""Test errors raised by func_serialize."""
        with pytest.raises(TypeError):
            instance.func_serialize(None)

    def test_deserialize_no_header(self, instance, testing_options,
                                   python_class, map_sent2recv, options):
        r"""Test deserialization of frame output without a header."""
        if options.get('no_header', False):
            return
        kws = instance.get_testing_options(no_header=True)
        kws['kwargs'].pop('no_header', None)
        no_head_inst = python_class(**kws['kwargs'])
        x = no_head_inst.serialize(kws['objects'][0])
        y = instance.deserialize(x)[0]
        assert(y == map_sent2recv(testing_options['objects'][0]))



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
import copy
import numpy as np
from yggdrasil import units, platform, constants


class TestScalarMetaschemaType(base_class):
    r"""Test class for ScalarMetaschemaType class with float."""
    
    _mod = 'yggdrasil.metaschema.datatypes.ScalarMetaschemaType'
    _cls = 'ScalarMetaschemaType'

    @pytest.fixture(scope="class", params=[False, True])
    def explicit(self, request):
        r"""bool: If True the type is explicit."""
        return request.param

    @pytest.fixture(scope="class")
    def class_name(self, explicit, subtype):
        r"""Name of class that will be tested."""
        if not explicit:
            return self._cls
        return f"{subtype.title()}MetaschemaType"

    @pytest.fixture(scope="class")
    def precision(self, subtype, array_contents, explicit):
        r"""int: Precision to test."""
        if subtype == 'complex':
            return 64
        elif subtype in ['bytes', 'unicode']:
            max_len = len(max(array_contents, key=len))
            if subtype == 'unicode':
                return max_len * 32
            else:
                return max_len * 8
        elif explicit and (subtype == 'float'):
            return 64
        return 32

    @pytest.fixture(scope="class", params=list(constants.VALID_TYPES.keys()))
    def subtype(self, request):
        r"""str: Scalar base type."""
        return request.param

    @pytest.fixture(scope="class")
    def shape(self):
        r"""int,tuple: Shape of scalar/array."""
        return 1

    @pytest.fixture(scope="class")
    def value_units(self):
        r"""str: Units that should be added to the type definition."""
        return None

    @pytest.fixture(scope="class")
    def valid_units(self):
        r"""list: Units that are compatible with the typedef units."""
        return []

    @pytest.fixture(scope="class")
    def invalid_units(self):
        r"""list: Units that are NOT compatible with the typedef units."""
        return []

    @pytest.fixture(scope="class")
    def array_contents(self, subtype):
        r"""Array contents."""
        if subtype in ['bytes', 'unicode']:
            return ['one', 'two', 'three']
        return None
    
    @pytest.fixture(scope="class")
    def typedef_base(self, explicit, subtype, value_units):
        r"""dict: Base type definition."""
        out = {}
        if not explicit:
            out['subtype'] = subtype
        if value_units:
            out['units'] = value_units
        return out

    @pytest.fixture(scope="class")
    def dtype(self, subtype, precision):
        r"""str: Numpy Data type."""
        if subtype == 'bytes':
            return 'S%d' % (precision // 8)
        elif subtype == 'unicode':
            return 'U%d' % (precision // 32)
        else:
            return '%s%d' % (subtype, precision)

    @pytest.fixture(scope="class")
    def dtype_invalid(self, subtype):
        r"""str: Invalid numpy data type."""
        if subtype in ['bytes', 'unicode']:
            return 'float'
        else:
            return 'S10'

    @pytest.fixture(scope="class")
    def array(self, shape, dtype, array_contents):
        r"""np.ndarray: Array for testing."""
        if array_contents is None:
            return np.ones(shape, dtype)
        else:
            return np.array(array_contents, dtype)

    @pytest.fixture(scope="class")
    def invalid_array(self, shape, dtype_invalid):
        r"""np.ndarray: Invalid array."""
        return np.ones(shape, dtype_invalid)

    @pytest.fixture(scope="class")
    def value(self, class_name, array, value_units):
        r"""dict: Test value."""
        if 'Array' not in class_name:
            out = array[0]
        else:
            out = array
        if value_units:
            out = units.add_units(out, value_units)
        return out

    @pytest.fixture(scope="class")
    def invalid_decoded(self, class_name, array, invalid_array, dtype):
        r"""list: Objects that are invalid under this type."""
        out = []
        if 'Array' not in class_name:
            out += [array, invalid_array[0]]
        else:
            if array.ndim == 1:
                out += [array[0], np.ones((3, 4), dtype)]
            else:
                out += [array[0][0], array[0]]
            out.append(invalid_array)
        return out

    @pytest.fixture(scope="class")
    def valid_encoded(self, python_class, precision, value, explicit,
                      subtype, valid_units, shape):
        r"""list: Encoded objects that are valid under this type."""
        out = [{'type': python_class.name,
                'precision': precision,
                'units': '',
                'data': value.tobytes()}]
        if not explicit:
            out[0]['subtype'] = subtype
        for x in valid_units:
            out.append(dict(out[0], units=x))
        if isinstance(shape, tuple):
            out[0]['shape'] = shape
        elif shape > 1:
            out[0]['length'] = shape
        return out

    @pytest.fixture(scope="class")
    def valid_decoded(self, value, valid_units):
        r"""list: Objects that are valid under this type."""
        out = [value]
        for x in valid_units:
            out.append(units.add_units(copy.deepcopy(out[0]), x))
        return out
    
    @pytest.fixture(scope="class")
    def precision_value(self, class_name, subtype, precision, array):
        r"""np.ndarray: Value with specific precision."""
        if subtype == 'bytes':
            new_dtype = 'S%d' % (precision * 2 // 8)
        elif subtype == 'unicode':
            new_dtype = 'U%d' % (precision * 2 // 32)
        else:
            new_dtype = '%s%d' % (subtype, precision * 2)
        if platform._is_win and (new_dtype == 'float128'):  # pragma: windows
            return None
        else:
            prec_array = array.astype(new_dtype)
            if 'Array' not in class_name:
                return prec_array[0]
            else:
                return prec_array

    @pytest.fixture(scope="class")
    def compatible_objects(self, subtype, precision, explicit, value,
                           precision_value):
        r"""list: Objects that are compatible with this type."""
        out = [(value, value, None)]
        if precision_value is not None:
            if not explicit:
                out.append(
                    (value, precision_value, {'subtype': subtype,
                                              'precision': precision * 2}))
            else:
                out.append(
                    (value, precision_value, {'precision': precision * 2}))
        return out
            
    @pytest.fixture(scope="class")
    def valid_normalize(self, class_name, explicit, subtype, value):
        r"""list: Pairs of pre-/post-normalized objects."""
        if 'Array' not in class_name:
            if explicit:
                if subtype == 'bytes':
                    return [(1, b'1'), (u'1', b'1')]
                elif subtype == 'unicode':
                    return [(1, u'1'), (b'1', u'1')]
                else:
                    return [(str(value), value), ('hello', 'hello')]
        return []

    @pytest.fixture(scope="class")
    def invalid_encoded(self, class_name, explicit, invalid_units):
        r"""list: Encoded objects that are invalid under this type."""
        out = [{}]
        if explicit and ('Array' not in class_name):
            out.append({'type': 'scalar', 'subtype': 'invalid'})
        for x in invalid_units:
            out.append(dict(out[0], units=x))
        return out
            
    @pytest.fixture(scope="class")
    def invalid_validate(self):
        r"""list: Objects that are invalid under this type."""
        return [None, np.array([None, 1, list()], dtype=object)]

    def test_from_array(self, value, typedef_base, array, instance,
                        nested_approx):
        r"""Test getting object from array."""
        test_val = value
        test_kws = {}
        if 'units' in typedef_base:
            test_val = units.add_units(test_val, typedef_base['units'])
            test_kws['unit_str'] = typedef_base['units']
        assert(instance.from_array(array, **test_kws) == nested_approx(test_val))


class TestScalarMetaschemaType_prec(TestScalarMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with precision."""

    @pytest.fixture(scope="class")
    def explicit(self):
        r"""bool: If True the type is explicit."""
        return False
    
    @pytest.fixture(scope="class")
    def subtype(self):
        r"""str: Scalar base type."""
        return "float"
    
    @pytest.fixture(scope="class")
    def typedef_base(self, explicit, subtype, precision):
        r"""dict: Base type definition."""
        return {'subtype': subtype,
                'precision': precision}
    
    @pytest.fixture(scope="class")
    def valid_encoded(self, python_class, precision, value, explicit,
                      subtype):
        r"""list: Encoded objects that are valid under this type."""
        out = [{'type': python_class.name,
                'precision': precision,
                'units': '',
                'data': value.tobytes(),
                'subtype': subtype}]
        # out.append(copy.deepcopy(out[0]))
        return out
    
    @pytest.fixture(scope="class")
    def invalid_encoded(self, precision, valid_encoded, precision_value):
        r"""list: Encoded objects that are invalid under this type."""
        out = [{'precision': precision / 2}]  # compatible precision
        out.append(copy.deepcopy(valid_encoded[0]))
        if precision_value is not None:
            out[-1]['precision'] = precision * 2
        return out
    
    @pytest.fixture(scope="class")
    def invalid_decoded(self, class_name, array, invalid_array, dtype,
                        precision_value):
        r"""list: Objects that are invalid under this type."""
        out = []
        if 'Array' not in class_name:
            out += [array, invalid_array[0]]
        else:
            if array.ndim == 1:
                out += [array[0], np.ones((3, 4), dtype)]
            else:
                out += [array[0][0], array[0]]
            out.append(invalid_array)
        if precision_value is not None:
            out.append(precision_value)
        return out


class TestScalarMetaschemaType_units(TestScalarMetaschemaType):
    r"""Test class for ScalarMetaschemaType class with units."""

    @pytest.fixture(scope="class")
    def explicit(self):
        r"""bool: If True the type is explicit."""
        return False
    
    @pytest.fixture(scope="class")
    def subtype(self):
        r"""str: Scalar base type."""
        return "float"
    
    @pytest.fixture(scope="class")
    def value_units(self, valid_units):
        r"""str: Units that should be added to the type definition."""
        return valid_units[0]
    
    @pytest.fixture(scope="class")
    def valid_units(self):
        r"""list: Units that are compatible with the typedef units."""
        return ['cm', 'm']

    @pytest.fixture(scope="class")
    def invalid_units(self):
        r"""list: Invalid units."""
        return ['s']



import os
import uuid
import copy
import pytest
from yggdrasil.communication import new_comm, get_comm
from yggdrasil.tools import get_supported_comm
from tests import TestComponentBase


def test_registry():
    r"""Test registry of comm."""
    from yggdrasil.communication import CommBase
    comm_class = 'CommBase'
    key = 'key1'
    value = None
    assert(not CommBase.is_registered(comm_class, key))
    assert(not CommBase.unregister_comm(comm_class, key))
    assert(CommBase.get_comm_registry(None) == {})
    assert(CommBase.get_comm_registry(comm_class) == {})
    CommBase.register_comm(comm_class, key, value)
    assert(key in CommBase.get_comm_registry(comm_class))
    assert(CommBase.is_registered(comm_class, key))
    assert(not CommBase.unregister_comm(comm_class, key, dont_close=True))
    CommBase.register_comm(comm_class, key, value)
    assert(not CommBase.unregister_comm(comm_class, key))


_communicators = sorted([x for x in get_supported_comm()
                         if x not in ['mpi', 'value', 'rest', 'rmq_async']])


class BaseComm(TestComponentBase):

    _component_type = 'comm'
    parametrize_commtype = _communicators
    parametrize_use_async = [False, True]

    @pytest.fixture(scope="class", autouse=True)
    def component_subtype(self, commtype):
        r"""Subtype of component being tested."""
        return commtype

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self, request):
        r"""Communicator type being tested."""
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def use_async(self, request):
        r"""Whether communicator should be asynchronous or not."""
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def check_installation(self, python_class):
        r"""Check that the communicator is installed."""
        if not python_class.is_installed(language="python"):
            pytest.skip("communicator not installed")

    @pytest.fixture(scope="class")
    def global_name(self, commtype, global_uuid):
        r"""Unique name for the test communicators."""
        return commtype + global_uuid

    @pytest.fixture
    def name(self, commtype, uuid):
        r"""Unique name for the test communicators."""
        return commtype + uuid

    @pytest.fixture(scope="class")
    def sleep_after_connect(self):
        r"""Indicates if sleep should occur after comm creation."""
        return False

    def get_send_comm_kwargs(self, commtype, use_async,
                             testing_options, **kwargs):
        r"""Get keyword arguments for creating a send comm."""
        kws = dict(commtype=commtype, reverse_names=True,
                   direction='send', use_async=use_async)
        kws.update(testing_options['kwargs'])
        kws.update(kwargs)
        return kws

    def get_recv_comm_kwargs(self, commtype, send_comm,
                             testing_options, **kwargs):
        r"""Get keyword arguments for creating a recv comm."""
        kws = dict(send_comm.opp_comm_kwargs(), commtype=commtype)
        kws.update(testing_options.get('recv_kwargs', {}))
        kws.update(kwargs)
        return kws

    def create_send_comm(self, name, commtype, use_async, testing_options,
                         **kwargs):
        r"""Create a send comm."""
        kws = self.get_send_comm_kwargs(commtype, use_async,
                                        testing_options, **kwargs)
        x = new_comm(name, **kws)
        assert(x.is_open)
        return x

    def create_recv_comm(self, name, commtype, send_comm, testing_options,
                         **kwargs):
        r"""Create a receive communicator."""
        kws = self.get_recv_comm_kwargs(commtype, send_comm,
                                        testing_options, **kwargs)
        x = get_comm(name, **kws)
        assert(x.is_open)
        x.drain_server_signon_messages()
        return x

    @pytest.fixture(scope="class")
    def process_send_message(self):
        r"""Factory for method to finalize messages for sending."""
        def wrapped_process_send_message(obj):
            return obj
        return wrapped_process_send_message

    @pytest.fixture(scope="class")
    def map_sent2recv(self):
        r"""Factory for method to convert sent messages to received."""
        def wrapped_map_sent2recv(obj):
            return obj
        return wrapped_map_sent2recv

    @pytest.fixture(scope="class")
    def n_msg_expected(self):
        r"""Number of expected messages."""
        return 1
    
    @pytest.fixture(scope="class")
    def do_send_recv(self, wait_on_function, testing_options, map_sent2recv,
                     n_msg_expected, nested_approx, logger, timeout):
        r"""Factory for method to perform send/recv checks for comms."""
        def wrapped_do_send_recv(send_comm, recv_comm, message=None,
                                 send_params=None, recv_params=None):
            if send_params is None:
                send_params = {}
            if recv_params is None:
                recv_params = {}
            assert(send_comm.n_msg_send == send_params.get('n_init', 0))
            assert(recv_comm.n_msg_recv == recv_params.get('n_init', 0))
            send_comm.printStatus(level='debug')
            recv_comm.printStatus(level='debug')
            # Send message
            if send_params.get('method', 'send') == 'send_eof':
                message = send_comm.eof_msg
                args = tuple([])
            else:
                if message is None:
                    message = testing_options['msg']
                args = (copy.deepcopy(message),)
            logger.debug(f"sending {send_params.get('count', 1)} "
                         f"copies of {message!s:.100}")
            for _ in range(send_params.get('count', 1)):
                flag = getattr(send_comm, send_params.get('method', 'send'))(
                    *args, **send_params.get('kwargs', {}))
                assert(flag == send_params.get('flag', True))
            # Receive message
            if (((not recv_params.get('skip_wait', False))
                 and ('eof' not in send_params.get('method', 'send')))):
                recv_params.setdefault('count', n_msg_expected)
            if 'message' not in recv_params:
                if 'eof' in send_params.get('method', 'send'):
                    recv_params['message'] = message
                else:
                    recv_params['message'] = map_sent2recv(message)
            logger.debug(f"expecting {recv_params.get('count', 1)} "
                         f"copies of {recv_params['message']!s:.100}")
            for _ in range(recv_params.get('count', 1)):
                if not recv_params.get('skip_wait', False):
                    wait_on_function(
                        lambda: (recv_comm.is_closed
                                 or (recv_comm.n_msg_recv > 0)))
                flag, msg = getattr(
                    recv_comm, recv_params.get('method', 'recv'))(
                        **recv_params.get('kwargs', {'timeout': 0}))
                assert(flag == recv_params.get('flag', True))
                assert(msg == nested_approx(recv_params['message']))
            if not send_params.get('skip_wait', False):
                wait_on_function(
                    lambda: send_comm.is_closed or (send_comm.n_msg_send == 0))
            if 'eof' not in send_params.get('method', 'send'):
                send_comm.wait_for_confirm(timeout=timeout)
                recv_comm.wait_for_confirm(timeout=timeout)
                assert(send_comm.is_confirmed)
                assert(recv_comm.is_confirmed)
                send_comm.confirm(noblock=True)
                recv_comm.confirm(noblock=True)
            assert(send_comm.n_msg_send == 0)
            assert(recv_comm.n_msg_recv == 0)
        return wrapped_do_send_recv

    @pytest.fixture(scope="class")
    def global_send_comm(self, global_name, commtype, use_async,
                         testing_options, close_comm):
        r"""Communicator for sending messages."""
        send_comm = self.create_send_comm(global_name, commtype, use_async,
                                          testing_options)
        yield send_comm
        send_comm.cleanup_comms()
        assert(len(send_comm.comm_registry()) == 0)
        close_comm(send_comm)

    @pytest.fixture(scope="class")
    def global_recv_comm(self, global_name, commtype, global_send_comm,
                         sleep_after_connect, testing_options, close_comm):
        r"""Communicator for receiving messages."""
        recv_comm = self.create_recv_comm(global_name, commtype,
                                          global_send_comm,
                                          testing_options)
        if sleep_after_connect:
            recv_comm.sleep()
        yield recv_comm
        close_comm(recv_comm)

    @pytest.fixture(scope="class")
    def global_comm(self, global_recv_comm):
        r"""Global communicator."""
        return global_recv_comm

    @pytest.fixture
    def send_comm(self, name, commtype, use_async, testing_options,
                  verify_count_threads, verify_count_comms,
                  verify_count_fds, close_comm):
        r"""Communicator for sending messages."""
        send_comm = self.create_send_comm(name, commtype, use_async,
                                          testing_options)
        yield send_comm
        close_comm(send_comm)

    @pytest.fixture
    def recv_comm(self, name, commtype, send_comm, testing_options,
                  sleep_after_connect, close_comm):
        r"""Communicator for receiving messages."""
        recv_comm = self.create_recv_comm(name, commtype, send_comm,
                                          testing_options)
        if sleep_after_connect:
            recv_comm.sleep()
        yield recv_comm
        close_comm(recv_comm)

    @pytest.fixture
    def maxMsgSize(self, global_comm):
        r"""int: Maximum message size."""
        return global_comm.maxMsgSize

    @pytest.fixture
    def msg_long(self, testing_options, maxMsgSize, process_send_message):
        r"""str: Large test message for sending."""
        out = testing_options['msg']
        if isinstance(out, bytes):
            out += (maxMsgSize * b'0')
        return process_send_message(out)

    @pytest.fixture(scope="class")
    def msg_array(self, testing_options, process_send_message, python_class):
        r"""array: Test message that should be used for send_array/recv_array
        tests."""
        out = testing_options.get('msg_array', None)
        if out is None:
            if python_class._commtype == 'ipc':
                out = python_class.get_testing_options(
                    table_example=True, array_columns=True,
                    include_oldkws=True)['msg_array']
            else:
                pytest.skip("No array message for communicator")
        return process_send_message(out)

    @pytest.fixture(scope="class")
    def msg_dict(self, testing_options, process_send_message, python_class):
        r"""dict: Test message that should be used for send_dict/recv_dict
        tests."""
        out = testing_options.get('dict', None)
        if not out:
            if python_class._commtype in ['ipc', 'fork']:
                out = python_class.get_testing_options(
                    table_example=True, array_columns=True)['dict']
            else:
                pytest.skip("No dict message for communicator")
        return process_send_message(out)

    @pytest.fixture(scope="class")
    def msg_filter_send(self, testing_options,
                        process_send_message):
        r"""object: Message to filter out on the send side."""
        objs = testing_options['objects']
        if len(objs) < 1:  # pragma: debug
            pytest.skip("There aren't enough objects.")
        return process_send_message(objs[0])

    @pytest.fixture(scope="class")
    def msg_filter_recv(self, testing_options,
                        process_send_message):
        r"""object: Message to filter out on the recv side."""
        objs = testing_options['objects']
        if (len(objs) >= 2):
            try:
                assert(objs[0] == objs[1])
            except BaseException:
                return process_send_message(objs[1])
        pytest.skip("There aren't enough unique objects.")

    @pytest.fixture(scope="class")
    def msg_filter_pass(self, testing_options,
                        process_send_message):
        r"""object: Message that won't be filtered out on send or recv."""
        objs = testing_options['objects']
        if len(objs) > 2:
            out = objs[2]
            assert(out != objs[0])
            assert(out != objs[1])
            return process_send_message(out)
        pytest.skip("There aren't enough unique objects.")

    @pytest.fixture
    def filtered_comms(self, send_comm, recv_comm, msg_filter_send,
                       msg_filter_recv, nested_approx):
        r"""Add filters to the send and receive communicators."""
        from yggdrasil.communication.filters.StatementFilter import (
            StatementFilter)
        from yggdrasil.communication.filters.FunctionFilter import (
            FunctionFilter)
        # Statement filter on send comm
        if isinstance(msg_filter_send, (str, bytes)):
            statement = '%x% != ' + repr(msg_filter_send)
        else:
            statement = 'repr(%x%) != r"""' + repr(msg_filter_send) + '"""'
        send_comm.filter = StatementFilter(statement=statement)
        # Function filter on recv comm

        def fcond(x):
            try:
                assert(x == nested_approx(msg_filter_recv))
                return False
            except BaseException:
                return True
        recv_comm.filter = FunctionFilter(function=fcond)
        yield
        send_comm.filter = None
        recv_comm.filter = None


@pytest.mark.suite("comms")
class TestComm(BaseComm):
    r"""Test communicator classes."""

    def test_empty_obj_recv(self, run_once, global_comm):
        r"""Test identification of empty message."""
        assert(global_comm.is_empty_recv(
            global_comm.empty_obj_recv))
        assert(not global_comm.is_empty_recv(global_comm.eof_msg))

    def test_error_name(self, python_class):
        r"""Test error on missing address."""
        with pytest.raises(RuntimeError):
            python_class('test%s' % uuid.uuid4())

    def test_error_send(self, monkeypatch, send_comm, testing_options,
                        magic_error_replacement):
        r"""Test error on send."""
        monkeypatch.setattr(send_comm, '_safe_send',
                            magic_error_replacement)
        flag = send_comm.send(testing_options['msg'])
        assert(not flag)

    def test_error_recv(self, monkeypatch, recv_comm,
                        magic_error_replacement):
        r"""Test error on recv."""
        monkeypatch.setattr(recv_comm, '_safe_recv',
                            magic_error_replacement)
        flag, msg_recv = recv_comm.recv(timeout=5.0)
        assert(not flag)

    def test_send_recv_after_close(self, commtype, send_comm, recv_comm,
                                   testing_options, wait_on_function):
        r"""Test that opening twice dosn't cause errors and that send/recv
        after close returns false."""
        send_comm.open()
        recv_comm.open()
        send_comm.close()
        recv_comm.close()
        assert(send_comm.is_closed)
        assert(recv_comm.is_closed)
        flag = send_comm.send(testing_options['msg'])
        assert(not flag)
        flag, msg_recv = recv_comm.recv()
        assert(not flag)

    def test_attributes(self, python_class, global_comm, testing_options):
        r"""Assert that the instance has all of the required attributes."""
        for a in testing_options.get('attributes', []):
            assert(hasattr(global_comm, a))
            getattr(global_comm, a)
        global_comm.opp_comm_kwargs()
        assert(not python_class.is_installed(language='invalid'))

    def test_invalid_direction(self, run_once, name, commtype, use_async,
                               testing_options):
        r"""Check that error raised for invalid direction."""
        kwargs = copy.deepcopy(testing_options)
        kwargs["kwargs"]["direction"] = "invalid"
        with pytest.raises(ValueError):
            self.create_send_comm(name, commtype, use_async, kwargs)

    def test_work_comm(self, uuid, recv_comm, testing_options, timeout):
        r"""Test creating/removing a work comm."""
        wc_send = recv_comm.create_work_comm()
        with pytest.raises(KeyError):
            recv_comm.add_work_comm(wc_send)
        # Create recv instance in way that tests new_comm
        header_recv = dict(id=uuid + '1', address=wc_send.address)
        recv_kwargs = recv_comm.get_work_comm_kwargs
        recv_kwargs.pop('async_recv_kwargs', None)
        recv_kwargs['work_comm_name'] = 'test_worker_%s' % header_recv['id']
        recv_kwargs['commtype'] = wc_send._commtype
        if isinstance(wc_send.opp_address, str):
            os.environ[recv_kwargs['work_comm_name']] = wc_send.opp_address
        else:
            recv_kwargs['address'] = wc_send.opp_address
        wc_recv = recv_comm.create_work_comm(**recv_kwargs)
        assert(wc_send.send(testing_options['msg']))
        flag, msg_recv = wc_recv.recv(timeout)
        assert(flag)
        assert(msg_recv == testing_options['msg'])
        # Assert errors on second attempt
        with pytest.raises(RuntimeError):
            wc_recv.recv()
        recv_comm.remove_work_comm(wc_send.uuid)
        recv_comm.remove_work_comm(wc_recv.uuid)
        recv_comm.remove_work_comm(wc_recv.uuid)
        # Create work comm that should be cleaned up on teardown
        recv_comm.create_work_comm()

    def test_drain_messages(self, global_send_comm, global_recv_comm,
                            timeout):
        r"""Test waiting for messages to drain."""
        global_send_comm.drain_messages(timeout=timeout)
        assert(global_send_comm.n_msg_send_drain == 0)
        if not global_recv_comm.is_file:
            global_recv_comm.drain_messages(timeout=timeout)
            assert(global_recv_comm.n_msg_recv_drain == 0)
        with pytest.raises(ValueError):
            global_send_comm.drain_messages(variable='n_msg_invalid')
        with pytest.raises(ValueError):
            global_recv_comm.drain_messages(variable='n_msg_invalid')

    def test_recv_nomsg(self, global_recv_comm, polling_interval):
        r"""Test recieve when there is no waiting message."""
        flag, msg_recv = global_recv_comm.recv(timeout=polling_interval)
        assert(flag)
        assert(not msg_recv)

    def test_send_recv(self, send_comm, recv_comm, do_send_recv):
        r"""Test send/recv of a small message."""
        do_send_recv(send_comm, recv_comm)

    def test_send_recv_eof(self, send_comm, recv_comm, do_send_recv,
                           timeout):
        r"""Test send/recv of EOF message."""
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'},
                     recv_params={'flag': False, 'skip_wait': True,
                                  'kwargs': {'timeout': timeout}})

    def test_send_recv_eof_no_close(self, send_comm, recv_comm, do_send_recv):
        r"""Test send/recv of EOF message with no close."""
        if recv_comm is not None:
            recv_comm.close_on_eof_recv = False
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'})

    def test_send_recv_nolimit(self, msg_long, send_comm, recv_comm,
                               do_send_recv):
        r"""Test send/recv of a large message."""
        do_send_recv(send_comm, recv_comm, msg_long,
                     send_params={'method': 'send_nolimit',
                                  'kwargs': {
                                      'header_kwargs': {'x': msg_long}}},
                     recv_params={'method': 'recv_nolimit'})
        if send_comm is not None:
            send_comm.printStatus()
            send_comm.printStatus(return_str=True)
        if recv_comm is not None:
            recv_comm.printStatus()
            recv_comm.printStatus(return_str=True)
        if send_comm is not None:
            send_comm.linger()

    def test_send_recv_filter_eof(self, run_once, filtered_comms, send_comm,
                                  recv_comm, do_send_recv):
        r"""Test send/recv of EOF with filter."""
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'},
                     recv_params={'flag': False})
        if recv_comm is not None:
            assert(recv_comm.is_closed)

    def test_send_recv_filter_pass(self, filtered_comms, msg_filter_pass,
                                   send_comm, recv_comm, do_send_recv):
        r"""Test send/recv with filter that passes both messages."""
        do_send_recv(send_comm, recv_comm, msg_filter_pass)

    def test_send_recv_filter_send_filter(self, filtered_comms,
                                          msg_filter_send, send_comm,
                                          recv_comm, polling_interval,
                                          do_send_recv):
        r"""Test send/recv with filter that blocks send."""
        if recv_comm is None:
            empty_msg = send_comm.empty_obj_recv
        else:
            empty_msg = recv_comm.empty_obj_recv
        do_send_recv(send_comm, recv_comm, msg_filter_send,
                     recv_params={'message': empty_msg,
                                  'skip_wait': True,
                                  'kwargs': {'timeout': polling_interval}})
        
    def test_send_recv_filter_recv_filter(self, filtered_comms,
                                          msg_filter_recv, send_comm,
                                          recv_comm, polling_interval,
                                          do_send_recv):
        r"""Test send/recv with filter that blocks recv."""
        # Wait if not async?
        if recv_comm is None:
            empty_msg = send_comm.empty_obj_recv
        else:
            empty_msg = recv_comm.empty_obj_recv
        do_send_recv(send_comm, recv_comm, msg_filter_recv,
                     recv_params={'message': empty_msg,
                                  'skip_wait': True,
                                  'kwargs': {'timeout': 10 * polling_interval}})

    def test_send_recv_raw(self, send_comm, recv_comm, testing_options):
        r"""Test send/recv of a small message."""
        from yggdrasil.communication import CommBase

        def dummy(msg):
            msg.msg = b''
            msg.args = b''
            msg.length = 0
            return msg
        assert(send_comm.send(testing_options['msg']))
        msg = recv_comm.recv(
            timeout=60.0, skip_deserialization=True,
            return_message_object=True, after_finalize_message=[dummy])
        assert(msg.finalized)
        assert(recv_comm.finalize_message(msg) == msg)
        msg.finalized = False
        assert(recv_comm.is_empty_recv(msg.args))
        msg = recv_comm.finalize_message(msg)
        assert(msg.flag == CommBase.FLAG_EMPTY)

    def test_send_recv_array(self, run_once, send_comm, recv_comm, msg_array,
                             do_send_recv):
        r"""Test send/recv of a array message."""
        do_send_recv(send_comm, recv_comm, msg_array,
                     send_params={'method': 'send_array'},
                     recv_params={'method': 'recv_array'})

    def test_send_recv_dict(self, send_comm, recv_comm, msg_dict,
                            do_send_recv):
        r"""Test send/recv message as dict."""
        do_send_recv(send_comm, recv_comm, msg_dict,
                     send_params={'method': 'send_dict'},
                     recv_params={'method': 'recv_dict'})
        
    def test_send_recv_dict_names(self, run_once, send_comm, recv_comm,
                                  msg_dict, testing_options,
                                  do_send_recv):
        field_order = testing_options.get('field_names', None)
        if not field_order:
            pytest.skip("No field_order option for communicator")
        do_send_recv(send_comm, recv_comm, msg_dict,
                     send_params={'method': 'send_dict',
                                  'kwargs': {'field_order': field_order}},
                     recv_params={'method': 'recv_dict',
                                  'kwargs': {'field_order': field_order}})

    def test_purge(self, send_comm, recv_comm, testing_options,
                   wait_on_function, n_msg_expected):
        r"""Test purging messages from the comm."""
        assert(send_comm.n_msg == 0)
        assert(recv_comm.n_msg == 0)
        if send_comm.is_async:
            assert(send_comm.n_msg_direct == 0)
        if recv_comm.is_async:
            assert(recv_comm.n_msg_direct == 0)
        # Purge recv while open
        flag = send_comm.send(testing_options['msg'])
        assert(flag)
        wait_on_function(lambda: recv_comm.n_msg == n_msg_expected)
        recv_comm.purge()
        # Uni-directional comms can't know about messages sent
        # assert(send_comm.n_msg == 0)
        assert(recv_comm.n_msg == 0)
        # Purge recv while closed
        recv_comm.close()
        recv_comm.purge()



import pytest
import os
import pprint
import tempfile
import subprocess
from jsonschema import ValidationError
from yggdrasil import schema, components


def filter_func_ex():  # pragma: no cover
    r"""Test function for normalizing filters."""
    return False


filter_func_ex_name = '%s:filter_func_ex' % __file__


@pytest.fixture
def normalize_objects(patch_equality, functions_equality):
    filter_func_ex2 = patch_equality(filter_func_ex, functions_equality)
    return [
        ({'models': [{
            'name': 'modelA',
            'language': 'c',
            'args': 'model.c',
            'outputs': [
                {'name': 'outputA',
                 'column_names': ['a', 'b'],
                 'column_units': ['cm', 'g'],
                 'filter': {
                     'function': filter_func_ex_name}}],
            'working_dir': os.getcwd()}],
          'connections': [{
              'inputs': 'outputA',
              'outputs': 'fileA.txt',
              'seritype': 'direct',
              'working_dir': os.getcwd()}]},
         {'models': [{
             'name': 'modelA',
             'language': 'c',
             'args': ['model.c'],
             'inputs': [{'commtype': 'default',
                         'datatype': {'type': 'bytes'},
                         'is_default': True,
                         'name': 'modelA:input'}],
             'outputs': [{'name': 'modelA:outputA',
                          'commtype': 'default',
                          'datatype': {'type': 'bytes'},
                          'filter': {
                              'function': filter_func_ex2}}],
             'working_dir': os.getcwd()}],
          'connections': [{
              'inputs': [
                  {'name': 'modelA:outputA',
                   'datatype': {'type': 'bytes'},
                   'commtype': 'default',
                   'filter': {
                       'function': filter_func_ex2}}],
              'outputs': [
                  {'name': 'fileA.txt',
                   'filetype': 'binary',
                   'working_dir': os.getcwd(),
                   'serializer': {'seritype': 'direct'},
                   'field_names': ['a', 'b'],
                   'field_units': ['cm', 'g']}]}]})]


def test_get_json_schema():
    r"""Test getting pure JSON version of schema."""
    test_file = 'strict_json_schema.json'
    schema.get_json_schema(test_file)
    assert(os.path.isfile(test_file))
    os.remove(test_file)
    

def test_SchemaRegistry():
    r"""Test schema registry."""
    with pytest.raises(ValueError):
        schema.SchemaRegistry({})
    x = schema.SchemaRegistry()
    assert((x == 0) is False)
    fname = os.path.join(tempfile.gettempdir(), 'temp.yml')
    with open(fname, 'w') as fd:
        fd.write('')
    with pytest.raises(Exception):
        x.load(fname)
    os.remove(fname)
    

def test_default_schema():
    r"""Test getting default schema."""
    s = schema.get_schema()
    assert(s is not None)
    schema.clear_schema()
    assert(schema._schema is None)
    s = schema.get_schema()
    assert(s is not None)
    for k in s.keys():
        assert(isinstance(s[k].subtypes, list))
        assert(isinstance(s[k].classes, list))
        for ksub in s[k].classes:
            s[k].get_subtype_properties(ksub)
            s[k].default_subtype
    s.get_schema(relaxed=True)
    s.get_schema(allow_instance=True)
    s.definitions
    s.form_schema


def test_create_schema():
    r"""Test re-creating the schema."""
    f_schema = schema._schema_fname
    f_consts = os.path.join(os.path.dirname(schema.__file__), 'constants.py')
    old_schema = open(f_schema, 'r').read()
    old_consts = open(f_consts, 'r').read()
    try:
        os.remove(f_schema)
        open(f_consts, 'w').write(
            old_consts.split(schema._constants_separator)[0]
            + schema._constants_separator)
        subprocess.check_call(['yggschema'])
        new_schema = open(f_schema, 'r').read()
        new_consts = open(f_consts, 'r').read()
        assert(new_consts == old_consts)
        assert(new_schema == old_schema)
    finally:
        open(f_schema, 'w').write(old_schema)
        open(f_consts, 'w').write(old_consts)


def test_save_load_schema():
    r"""Test saving & loading schema."""
    fname = 'test_schema.yml'
    if os.path.isfile(fname):  # pragma: debug
        os.remove(fname)
    # Test saving/loading schema
    s0 = schema.load_schema()
    s0.save(fname)
    assert(s0 is not None)
    assert(os.path.isfile(fname))
    s1 = schema.get_schema(fname)
    assert(s1.schema == s0.schema)
    # assert(s1 == s0)
    os.remove(fname)
    # Test getting schema
    s2 = schema.load_schema(fname)
    assert(os.path.isfile(fname))
    assert(s2.schema == s0.schema)
    assert(s2 == s0)
    os.remove(fname)


def test_cdriver2filetype_error():
    r"""Test errors in cdriver2filetype."""
    with pytest.raises(ValueError):
        schema.cdriver2filetype('invalid')


def test_standardize():
    r"""Test standardize."""
    vals = [(False, ['inputs', 'outputs'], ['_file'],
             {'input': 'inputA', 'output_file': 'outputA'},
             {'inputs': [{'name': 'inputA'}],
              'outputs': [{'name': 'outputA'}]}),
            (True, ['input', 'output'], ['_file'],
             {'inputs': 'inputA', 'output_files': 'outputA'},
             {'input': [{'name': 'inputA'}],
              'output': [{'name': 'outputA'}]})]
    for is_singular, keys, suffixes, x, y in vals:
        schema.standardize(x, keys, suffixes=suffixes, is_singular=is_singular)
        assert(x == y)


def test_normalize(normalize_objects):
    r"""Test normalization of legacy formats."""
    s = schema.get_schema()
    for x, y in normalize_objects:
        a = s.normalize(x, backwards_compat=True)  # , show_errors=True)
        try:
            assert(a == y)
        except BaseException:  # pragma: debug
            print("Unexpected Normalization:\n\nA:")
            pprint.pprint(a)
            print('\nB:')
            pprint.pprint(y)
            raise


def test_cdriver2commtype_error():
    r"""Test error when invalid driver supplied."""
    with pytest.raises(ValueError):
        schema.cdriver2commtype('invalid')


def test_get_schema_subtype():
    r"""Test get_schema_subtype for allow_instance=True."""
    component = 'serializer'
    subtype = 'direct'
    doc = {'seritype': subtype}
    valid = components.create_component(component, seritype=subtype)
    invalid = components.create_component(component, seritype='json')
    s = schema.get_schema()
    kwargs = {'subtype': subtype, 'allow_instance': True}
    s.validate_component(component, doc, **kwargs)
    s.validate_component(component, valid, **kwargs)
    with pytest.raises(ValidationError):
        s.validate_component(component, invalid, **kwargs)
    s.validate_component(component, doc, subtype=subtype)
    with pytest.raises(ValidationError):
        s.validate_component(component, valid, subtype=subtype)
    # Test for base
    s.validate_component(component, valid, subtype='base',
                         allow_instance=True)
    s.validate_component(component, invalid, subtype='base',
                         allow_instance=True)


def test_get_model_form_schema():
    r"""Test get_model_form_schema."""
    fname = os.path.join(tempfile.gettempdir(), 'temp.json')
    try:
        schema.get_model_form_schema(fname_dst=fname)
        assert(os.path.isfile(fname))
    finally:
        if os.path.isfile(fname):
            os.remove(fname)


def test_update_constants(project_dir):
    r"""Test script to update constants and check that they have not changed."""
    filename = os.path.join(project_dir, 'constants.py')
    with open(filename, 'r') as fd:
        old = fd.read()
    try:
        schema.update_constants()
        with open(filename, 'r') as fd:
            new = fd.read()
        assert(old == new)
    finally:
        with open(filename, 'w') as fd:
            fd.write(old)



import pytest
import os
import numpy as np
from yggdrasil import serialize
from tests.examples import TestExample as base_class


@pytest.mark.extra_example
class TestExampleAsciiIO(base_class):
    r"""Test the AsciiIO example."""

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "ascii_io"

    @pytest.fixture
    def input_file(self, yamldir):
        r"""str: Input file."""
        return os.path.join(yamldir, 'Input', 'input_file.txt')

    @pytest.fixture
    def input_table(self, yamldir):
        r"""str: Input table file."""
        return os.path.join(yamldir, 'Input', 'input_table.txt')
        
    @pytest.fixture
    def input_array(self, yamldir):
        r"""str: Input array file."""
        return os.path.join(yamldir, 'Input', 'input_array.txt')

    # @pytest.fixture
    # def input_files(self, input_file, input_table, input_array):
    #     r"""list Input files for the run."""
    #     return [input_file, input_table, input_array]
        
    @pytest.fixture
    def output_file(self, tempdir):
        r"""str: Output file for the run."""
        return os.path.join(tempdir, 'output_file.txt')

    @pytest.fixture
    def output_table(self, tempdir):
        r"""str: Output table for the run."""
        return os.path.join(tempdir, 'output_table.txt')

    @pytest.fixture
    def output_array(self, tempdir):
        r"""str: Output array for the run."""
        return os.path.join(tempdir, 'output_array.txt')

    @pytest.fixture
    def output_files(self, output_file, output_table, output_array):
        r"""list: Output files for the run."""
        return [output_file, output_table, output_array]

    @pytest.fixture
    def results(self, input_file, input_table, input_array):
        r"""list: Results that should be found in the output files."""
        assert(os.path.isfile(input_file))
        assert(os.path.isfile(input_table))
        assert(os.path.isfile(input_array))
        with open(input_file, 'r') as fd:
            icont = fd.read()
        with open(input_table, 'rb') as fd:
            iATT = serialize.table_to_array(fd.read(), comment='#')
        with open(input_array, 'rb') as fd:
            iATA = serialize.table_to_array(fd.read(), comment='#')
        return [icont,
                (self.check_table, iATT),
                (self.check_table, iATA)]

    def check_table(self, fname, iAT):
        r"""Assert that contents of input/output ascii tables are identical."""
        with open(fname, 'rb') as fd:
            oAT = serialize.table_to_array(fd.read(), comment='#')
        np.testing.assert_equal(oAT, iAT)



import pytest
from tests.metaschema.datatypes.test_JSONArrayMetaschemaType import (
    TestJSONArrayMetaschemaType as base_class)
import copy
import numpy as np


def test_coerce(nested_approx):
    r"""Test serialization of coerced types."""
    from yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType import (
        JSONObjectMetaschemaType)
    from yggdrasil import serialize
    typedef = {'type': 'object',
               'properties': {'a': {'type': '1darray',
                                    'subtype': 'float',
                                    'title': 'a',
                                    'precision': 64}}}
    x = JSONObjectMetaschemaType(**typedef)
    key_order = ['a']
    msg_recv = {'a': np.zeros(3, 'float64')}
    msg_send_list = [{'a': np.zeros(3, 'float32')},
                     serialize.dict2numpy(msg_recv, order=key_order),
                     serialize.dict2pandas(msg_recv, order=key_order),
                     serialize.dict2list(msg_recv, order=key_order)]

    def do_send_recv(msg_send):
        msg_seri = x.serialize(msg_send, tyepdef=typedef, key_order=key_order)
        assert(x.deserialize(msg_seri)[0] == nested_approx(msg_recv))

    for y in msg_send_list:
        do_send_recv(y)


class TestJSONObjectMetaschemaType(base_class):
    r"""Test class for JSONObjectMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.JSONObjectMetaschemaType'
    _cls = 'JSONObjectMetaschemaType'
    test_encode_data_readable = None
    test_container_errors = None
    test_item_dictionary = None
    test_validate_errors = None

    @pytest.fixture(scope="class")
    def keys(self):
        return 'abcdefg'

    @pytest.fixture(scope="class")
    def value(self, keys, container_values):
        r"""list: Test value."""
        return {k: v for k, v in zip(keys, container_values)}

    @pytest.fixture(scope="class")
    def fulldef(self, python_class, keys, container_definitions):
        r"""dict: Full type definitions."""
        return {'type': python_class.name,
                'properties': {k: v for k, v in
                               zip(keys, container_definitions)}}
    
    @pytest.fixture(scope="class")
    def typedef_base(self, keys, container_typedefs):
        r"""dict: Base type definition."""
        return {'properties': {k: v for k, v in
                               zip(keys, container_typedefs)}}

    @pytest.fixture(scope="class")
    def invalid_encoded(self, fulldef):
        r"""list: Encoded objects that are invalid under this type."""
        out = [{},
               {'type': fulldef['type'],
                'properties': {'a': fulldef['properties']['a']}},
               copy.deepcopy(fulldef)]
        del out[-1]['properties']['a']['type']
        out.append(copy.deepcopy(fulldef))
        out[-1]['properties']['a']['type'] = 'invalid'
        return out
    
    @pytest.fixture(scope="class")
    def valid_normalize(self):
        r"""list: Pairs of pre-/post-normalized objects."""
        return [(None, None)]
